"""
MeasureObjectNeighbors
======================

**MeasureObjectNeighbors** calculates how many neighbors each object
has and records various properties about the neighbors’ relationships,
including the percentage of an object’s edge pixels that touch a
neighbor. Please note that the distances reported for object
measurements are center-to-center distances, not edge-to-edge distances.

Given an image with objects identified (e.g., nuclei or cells), this
module determines how many neighbors each object has. You can specify
the distance within which objects should be considered neighbors, or
that objects are only considered neighbors if they are directly
touching.

|

============ ============ ===============
Supports 2D? Supports 3D? Respects masks?
============ ============ ===============
YES          YES          NO
============ ============ ===============

See also
^^^^^^^^

See also the **Identify** modules.

Measurements made by this module
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

**Object measurements**

-  *NumberOfNeighbors:* Number of neighbor objects.
-  *PercentTouching:* Percent of the object’s boundary pixels that touch
   neighbors, after the objects have been expanded to the specified
   distance.
-  *FirstClosestObjectNumber:* The index of the closest object.
-  *FirstClosestDistance:* The distance to the closest object (in units
   of pixels), measured between object centers.
-  *SecondClosestObjectNumber:* The index of the second closest object.
-  *SecondClosestDistance:* The distance to the second closest object (in units
   of pixels), measured between object centers.
-  *AngleBetweenNeighbors:* The angle formed with the object center as
   the vertex and the first and second closest object centers along the
   vectors.

**Object relationships:** The identity of the neighboring objects, for
each object. Since per-object output is one-to-one and neighbors
relationships are often many-to-one, they may be saved as a separate
file in **ExportToSpreadsheet** by selecting *Object relationships* from
the list of objects to export.

Technical notes
^^^^^^^^^^^^^^^

Objects discarded via modules such as **IdentifyPrimaryObjects** or
**IdentifySecondaryObjects** will still register as neighbors for the
purposes of accurate measurement. For instance, if an object touches a
single object and that object had been discarded, *NumberOfNeighbors*
will be positive, but there may not be a corresponding
*ClosestObjectNumber*. This can be disabled in module settings.

"""

import numpy
import scipy
import scipy.ndimage
import scipy.signal
import skimage.morphology
from centrosome.cpmorphology import centers_of_labels, strel_disk
from centrosome.cpmorphology import fixup_scipy_ndimage_result as fix
from centrosome.outline import outline

D_ADJACENT = "Adjacent"
D_EXPAND = "Expand until adjacent"
D_WITHIN = "Within a specified distance"
D_ALL = [D_ADJACENT, D_EXPAND, D_WITHIN]

M_NUMBER_OF_NEIGHBORS = "NumberOfNeighbors"
M_PERCENT_TOUCHING = "PercentTouching"
M_FIRST_CLOSEST_OBJECT_NUMBER = "FirstClosestObjectNumber"
M_FIRST_CLOSEST_DISTANCE = "FirstClosestDistance"
M_SECOND_CLOSEST_OBJECT_NUMBER = "SecondClosestObjectNumber"
M_SECOND_CLOSEST_DISTANCE = "SecondClosestDistance"
M_ANGLE_BETWEEN_NEIGHBORS = "AngleBetweenNeighbors"
M_ALL = [
    M_NUMBER_OF_NEIGHBORS,
    M_PERCENT_TOUCHING,
    M_FIRST_CLOSEST_OBJECT_NUMBER,
    M_FIRST_CLOSEST_DISTANCE,
    M_SECOND_CLOSEST_OBJECT_NUMBER,
    M_SECOND_CLOSEST_DISTANCE,
    M_ANGLE_BETWEEN_NEIGHBORS,
]

C_NEIGHBORS = "Neighbors"

S_EXPANDED = "Expanded"
S_ADJACENT = "Adjacent"


def measureobjectneighbors(
    masks1: numpy.ndarray,
    masks2: numpy.ndarray,
    distance_method: str = D_EXPAND,
    distance: int = 5,
) -> dict[str, list[float | int]]:
    """
    Calculate neighbors of objects based on different methods. Supports 3D masks.

    Parameters
    ----------
    masks1 : (array of integers) label masks to be used as reference
    masks2 : (array of integers) label masks to be used as potential neighbors
    distance_method : str
        There are several methods by which to determine whether objects are
        neighbors:

        -  *%(D_ADJACENT)s:* In this mode, two objects must have adjacent
           boundary pixels to be neighbors.
        -  *%(D_EXPAND)s:* The objects are expanded until all pixels on the
           object boundaries are touching another. Two objects are neighbors if
           any of their boundary pixels are adjacent after expansion.
        -  *%(D_WITHIN)s:* Each object is expanded by the number of pixels you
           specify. Two objects are neighbors if they have adjacent pixels after
           expansion. Note that *all* objects are expanded by this amount (e.g.,
           if this distance is set to 10, a pair of objects will count as
           neighbors if their edges are 20 pixels apart or closer).

        For *%(D_ADJACENT)s* and *%(D_EXPAND)s*, the
        *%(M_PERCENT_TOUCHING)s* measurement is the percentage of pixels on
        the boundary of an object that touch adjacent objects. For
        *%(D_WITHIN)s*, two objects are touching if any of their boundary
        pixels are adjacent after expansion and *%(M_PERCENT_TOUCHING)s*
        measures the percentage of boundary pixels of an *expanded* object that
        touch adjacent objects.
    distance : int
        Neighbor distance (used only when “D_WITHIN” is selected).
        The Neighbor distance is the number of pixels that each object is
        expanded for the neighbor calculation. Expanded objects that touch are
        considered neighbors.

    Returns
    -------
    List of dictionaries with overlap features.
    """
    dimensions = masks1.ndim
    # has_pixels = masks.any()
    dimensions = masks1.ndim
    # has_pixels = masks.any()
    labels = masks1
    neighbor_labels = masks2
    # neighbor_objects = workspace.object_set.get_objects(self.neighbors_name.value)
    # neighbor_labels = neighbor_objects.small_removed_segmented
    # # neighbor_kept_labels = neighbor_objects.segmented
    # assert isinstance(neighbor_objects, Objects)

    nobjects = numpy.max(labels)
    nneighbors = numpy.max(neighbor_labels)

    neighbor_count = numpy.zeros((nobjects,))
    pixel_count = numpy.zeros((nobjects,))
    first_object_number = numpy.zeros((nobjects,), int)
    second_object_number = numpy.zeros((nobjects,), int)
    first_x_vector = numpy.zeros((nobjects,))
    second_x_vector = numpy.zeros((nobjects,))
    first_y_vector = numpy.zeros((nobjects,))
    second_y_vector = numpy.zeros((nobjects,))
    angle = numpy.zeros((nobjects,))
    percent_touching = numpy.zeros((nobjects,))
    # expanded_labels = None

    if distance_method == D_EXPAND:
        # Find the i,j coordinates of the nearest foreground point
        # to every background point
        if dimensions == 2:
            i, j = scipy.ndimage.distance_transform_edt(
                labels == 0, return_distances=False, return_indices=True
            )
            # Assign each background pixel to the label of its nearest
            # foreground pixel. Assign label to label for foreground.
            labels = labels[i, j]
        else:
            k, i, j = scipy.ndimage.distance_transform_edt(
                labels == 0, return_distances=False, return_indices=True
            )
            labels = labels[k, i, j]
        # expanded_labels = labels  # for display
        distance = 1  # dilate once to make touching edges overlap
    # These scale assignments are never used
    # scale = S_EXPANDED
    # if neighbors_are_objects:
    #     neighbor_labels = labels.copy()
    elif distance_method == D_WITHIN:
        distance = distance
        # scale = str(distance)
    elif distance_method == D_ADJACENT:
        distance = 1
        # scale = S_ADJACENT
    else:
        raise ValueError("Unknown distance method: %s" % distance_method)
    # if nneighbors > (1 if neighbors_are_objects else 0):
    if nneighbors > 1:
        first_objects = []
        second_objects = []
        object_indexes = numpy.arange(nobjects, dtype=numpy.int32) + 1
        #
        # First, compute the first and second nearest neighbors,
        # and the angles between self and the first and second
        # nearest neighbors
        #
        ocenters = centers_of_labels(labels).transpose()
        ncenters = centers_of_labels(neighbor_labels).transpose()
        # This is not used even in original implementation
        # areas = fix(scipy.ndimage.sum(numpy.ones(labels.shape), labels, object_indexes))
        perimeter_outlines = outline(labels)
        perimeters = fix(
            scipy.ndimage.sum(
                numpy.ones(labels.shape), perimeter_outlines, object_indexes
            )
        )

        #
        # order[:,0] should be arange(nobjects)
        # order[:,1] should be the nearest neighbor
        # order[:,2] should be the next nearest neighbor
        #
        order = numpy.zeros((nobjects, min(nneighbors, 3)), dtype=numpy.uint32)
        j = numpy.arange(nneighbors)
        # (0, 1, 2) unless there are less than 3 neighbors
        partition_keys = tuple(range(min(nneighbors, 3)))
        for i in range(nobjects):
            dr = numpy.sqrt(
                (ocenters[i, 0] - ncenters[j, 0]) ** 2
                + (ocenters[i, 1] - ncenters[j, 1]) ** 2
            )
            order[i, :] = numpy.argpartition(dr, partition_keys)[:3]

        # first_neighbor = 1 if neighbors_are_objects else 0
        # neighbors are objects because
        first_neighbor = 1
        first_object_index = order[:, first_neighbor]
        first_x_vector = ncenters[first_object_index, 1] - ocenters[:, 1]
        first_y_vector = ncenters[first_object_index, 0] - ocenters[:, 0]
        if nneighbors > first_neighbor + 1:
            second_object_index = order[:, first_neighbor + 1]
            second_x_vector = ncenters[second_object_index, 1] - ocenters[:, 1]
            second_y_vector = ncenters[second_object_index, 0] - ocenters[:, 0]
            v1 = numpy.array((first_x_vector, first_y_vector))
            v2 = numpy.array((second_x_vector, second_y_vector))
            #
            # Project the unit vector v1 against the unit vector v2
            #
            dot = numpy.sum(v1 * v2, 0) / numpy.sqrt(
                numpy.sum(v1**2, 0) * numpy.sum(v2**2, 0)
            )
            angle = numpy.arccos(dot) * 180.0 / numpy.pi

        # Make the structuring element for dilation
        if dimensions == 2:
            strel = strel_disk(distance)
        else:
            strel = skimage.morphology.ball(distance)
        #
        # A little bigger one to enter into the border with a structure
        # that mimics the one used to create the outline
        #
        if dimensions == 2:
            strel_touching = strel_disk(distance + 0.5)
        else:
            strel_touching = skimage.morphology.ball(distance + 0.5)
        #
        # Get the extents for each object and calculate the patch
        # that excises the part of the image that is "distance"
        # away
        if dimensions == 2:
            i, j = numpy.mgrid[0 : labels.shape[0], 0 : labels.shape[1]]

            minimums_i, maximums_i, _, _ = scipy.ndimage.extrema(
                i, labels, object_indexes
            )
            minimums_j, maximums_j, _, _ = scipy.ndimage.extrema(
                j, labels, object_indexes
            )

            minimums_i = numpy.maximum(fix(minimums_i) - distance, 0).astype(int)
            maximums_i = numpy.minimum(
                fix(maximums_i) + distance + 1, labels.shape[0]
            ).astype(int)
            minimums_j = numpy.maximum(fix(minimums_j) - distance, 0).astype(int)
            maximums_j = numpy.minimum(
                fix(maximums_j) + distance + 1, labels.shape[1]
            ).astype(int)
        else:
            k, i, j = numpy.mgrid[
                0 : labels.shape[0], 0 : labels.shape[1], 0 : labels.shape[2]
            ]

            minimums_k, maximums_k, _, _ = scipy.ndimage.extrema(
                k, labels, object_indexes
            )
            minimums_i, maximums_i, _, _ = scipy.ndimage.extrema(
                i, labels, object_indexes
            )
            minimums_j, maximums_j, _, _ = scipy.ndimage.extrema(
                j, labels, object_indexes
            )

            minimums_k = numpy.maximum(fix(minimums_k) - distance, 0).astype(int)
            maximums_k = numpy.minimum(
                fix(maximums_k) + distance + 1, labels.shape[0]
            ).astype(int)
            minimums_i = numpy.maximum(fix(minimums_i) - distance, 0).astype(int)
            maximums_i = numpy.minimum(
                fix(maximums_i) + distance + 1, labels.shape[1]
            ).astype(int)
            minimums_j = numpy.maximum(fix(minimums_j) - distance, 0).astype(int)
            maximums_j = numpy.minimum(
                fix(maximums_j) + distance + 1, labels.shape[2]
            ).astype(int)
        #
        # Loop over all objects
        # Calculate which ones overlap "index"
        # Calculate how much overlap there is of others to "index"
        #

        # This originally linked labels to post-filter labels
        # TODO replace with simpler procedure
        _, object_numbers = relate_labels(labels, labels)
        _, neighbor_numbers = relate_labels(neighbor_labels, neighbor_labels)
        for object_number in object_numbers:
            if object_number == 0:
                #
                # No corresponding object in small-removed. This means
                # that the object has no pixels, e.g., not renumbered.
                #
                continue
            index = object_number - 1
            if dimensions == 2:
                patch = labels[
                    minimums_i[index] : maximums_i[index],
                    minimums_j[index] : maximums_j[index],
                ]
                npatch = neighbor_labels[
                    minimums_i[index] : maximums_i[index],
                    minimums_j[index] : maximums_j[index],
                ]
            else:
                patch = labels[
                    minimums_k[index] : maximums_k[index],
                    minimums_i[index] : maximums_i[index],
                    minimums_j[index] : maximums_j[index],
                ]
                npatch = neighbor_labels[
                    minimums_k[index] : maximums_k[index],
                    minimums_i[index] : maximums_i[index],
                    minimums_j[index] : maximums_j[index],
                ]

            #
            # Find the neighbors
            #
            patch_mask = patch == (index + 1)
            if distance <= 5:
                extended = scipy.ndimage.binary_dilation(patch_mask, strel)
            else:
                extended = (
                    scipy.signal.fftconvolve(patch_mask, strel, mode="same") > 0.5
                )
            neighbors = numpy.unique(npatch[extended])
            neighbors = neighbors[neighbors != 0]
            # if neighbors_are_objects:
            # neighbors = neighbors[neighbors != object_number]
            neighbors = neighbors[neighbors != object_number]
            nc = len(neighbors)
            neighbor_count[index] = nc
            if nc > 0:
                first_objects.append(numpy.ones(nc, int) * object_number)
                second_objects.append(neighbors)
            #
            # Find the # of overlapping pixels. Dilate the neighbors
            # and see how many pixels overlap our image. Use a 3x3
            # structuring element to expand the overlapping edge
            # into the perimeter.
            #
            if dimensions == 2:
                outline_patch = (
                    perimeter_outlines[
                        minimums_i[index] : maximums_i[index],
                        minimums_j[index] : maximums_j[index],
                    ]
                    == object_number
                )
            else:
                outline_patch = (
                    perimeter_outlines[
                        minimums_k[index] : maximums_k[index],
                        minimums_i[index] : maximums_i[index],
                        minimums_j[index] : maximums_j[index],
                    ]
                    == object_number
                )
            # if neighbors_are_objects:
            extendme = (patch != 0) & (patch != object_number)
            if distance <= 5:
                extended = scipy.ndimage.binary_dilation(extendme, strel_touching)
            else:
                extended = (
                    scipy.signal.fftconvolve(extendme, strel_touching, mode="same")
                    > 0.5
                )
            # else:
            #     if distance <= 5:
            #         extended = scipy.ndimage.binary_dilation(
            #             (npatch != 0), strel_touching
            #         )
            #     else:
            #         extended = (
            #             scipy.signal.fftconvolve(
            #                 (npatch != 0), strel_touching, mode="same"
            #             )
            #             > 0.5
            #         )
            overlap = numpy.sum(outline_patch & extended)
            pixel_count[index] = overlap
        if sum([len(x) for x in first_objects]) > 0:
            first_objects = numpy.hstack(first_objects)
            reverse_object_numbers = numpy.zeros(
                max(numpy.max(object_numbers), numpy.max(first_objects)) + 1, int
            )
            reverse_object_numbers[object_numbers] = (
                numpy.arange(len(object_numbers)) + 1
            )
            first_objects = reverse_object_numbers[first_objects]

            second_objects = numpy.hstack(second_objects)
            reverse_neighbor_numbers = numpy.zeros(
                max(numpy.max(neighbor_numbers), numpy.max(second_objects)) + 1, int
            )
            reverse_neighbor_numbers[neighbor_numbers] = (
                numpy.arange(len(neighbor_numbers)) + 1
            )
            second_objects = reverse_neighbor_numbers[second_objects]
            to_keep = (first_objects > 0) & (second_objects > 0)
            first_objects = first_objects[to_keep]
            second_objects = second_objects[to_keep]
        else:
            first_objects = numpy.zeros(0, int)
            second_objects = numpy.zeros(0, int)
        percent_touching = pixel_count * 100 / perimeters
        object_indexes = object_numbers - 1
        neighbor_indexes = neighbor_numbers - 1
        #
        # Have to recompute nearest
        #
        # first_object_number = numpy.zeros(nkept_objects, int)
        # second_object_number = numpy.zeros(nkept_objects, int)
        first_object_number = numpy.zeros(nobjects, int)
        second_object_number = numpy.zeros(nobjects, int)
        # if nkept_objects > (1 if neighbors_are_objects else 0):
        di = (
            ocenters[object_indexes[:, numpy.newaxis], 0]
            - ncenters[neighbor_indexes[numpy.newaxis, :], 0]
        )
        dj = (
            ocenters[object_indexes[:, numpy.newaxis], 1]
            - ncenters[neighbor_indexes[numpy.newaxis, :], 1]
        )
        distance_matrix = numpy.sqrt(di * di + dj * dj)
        # distance_matrix[~has_pixels, :] = numpy.inf
        # distance_matrix[:, ~neighbor_has_pixels] = numpy.inf
        #
        # order[:,0] should be arange(nobjects)
        # order[:,1] should be the nearest neighbor
        # order[:,2] should be the next nearest neighbor
        #
        order = numpy.lexsort([distance_matrix]).astype(first_object_number.dtype)
        # Outcomments are due to these conditions being assumed
        # if neighbors_are_objects:
        # first_object_number[has_pixels] = order[has_pixels, 1] + 1
        # TODO check conditions in which they should be 0 and 1 or 0 and 2
        first_object_number = order[:, 0] + 1
        if order.shape[1] > 1:
            second_object_number = order[:, 1] + 1
        # else:
        #     first_object_number[has_pixels] = order[has_pixels, 0] + 1
        #     if order.shape[1] > 1:
        #         second_object_number[has_pixels] = order[has_pixels, 1] + 1
    else:
        object_indexes = object_numbers - 1
        neighbor_indexes = neighbor_numbers - 1
        first_objects = numpy.zeros(0, int)
        second_objects = numpy.zeros(0, int)
    #
    # Now convert all measurements from the small-removed to
    # the final number set.
    #
    neighbor_count = neighbor_count[object_indexes]
    # neighbor_count[~has_pixels] = 0
    percent_touching = percent_touching[object_indexes]
    # percent_touching[~has_pixels] = 0
    first_x_vector = first_x_vector[object_indexes]
    second_x_vector = second_x_vector[object_indexes]
    first_y_vector = first_y_vector[object_indexes]
    second_y_vector = second_y_vector[object_indexes]
    angle = angle[object_indexes]
    #
    # Record the measurements
    #
    # assert isinstance(workspace, Workspace)
    # m = workspace.measurements
    # assert isinstance(m, Measurements)
    # image_set = workspace.image_set
    features_and_data = [
        (M_NUMBER_OF_NEIGHBORS, neighbor_count),
        (M_FIRST_CLOSEST_OBJECT_NUMBER, first_object_number),
        (
            M_FIRST_CLOSEST_DISTANCE,
            numpy.sqrt(first_x_vector**2 + first_y_vector**2),
        ),
        (M_SECOND_CLOSEST_OBJECT_NUMBER, second_object_number),
        (
            M_SECOND_CLOSEST_DISTANCE,
            numpy.sqrt(second_x_vector**2 + second_y_vector**2),
        ),
        (M_ANGLE_BETWEEN_NEIGHBORS, angle),
        (M_PERCENT_TOUCHING, percent_touching),
    ]
    result = {}
    for feature_name, data in features_and_data:
        result[get_measurement_name(feature_name, distance_method, distance)] = data
    # TODO add related measurements
    # if len(first_objects) > 0:
    #     m.add_relate_measurement(
    #         self.module_num,
    #         NEIGHBORS,
    #         self.object_name.value,
    #         self.object_name.value
    #         if neighbors_are_objects
    #         else self.neighbors_name.value,
    #         m.image_set_number * numpy.ones(first_objects.shape, int),
    #         first_objects,
    #         m.image_set_number * numpy.ones(second_objects.shape, int),
    #         second_objects,
    #     )

    # labels = kept_labels
    # neighbor_labels = neighbor_kept_labels
    return result


def get_measurement_name(feature: str, distance_method: str, distance: int = 5):
    if distance_method == D_EXPAND:
        scale = S_EXPANDED
    elif distance_method == D_WITHIN:
        scale = str(distance)
    elif distance_method == D_ADJACENT:
        scale = S_ADJACENT
    return "_".join((C_NEIGHBORS, feature, scale))


def find_label_overlaps(parent_labels: numpy.ndarray, child_labels: numpy.ndarray):
    """
    Produces a matrix in which each row is a parent and each column is a child.


    Ported and modified from https://github.com/cellprofiler/CellProfiler/blob/main/src/subpackages/library/cellprofiler_library/functions/segmentation.py#L654
    """

    parent_count = numpy.max(parent_labels)
    child_count = numpy.max(child_labels)

    #
    # Only look at points that are labeled in parent and child
    #
    not_zero = (parent_labels > 0) & (child_labels > 0)
    not_zero_count = numpy.sum(not_zero)

    #
    # each row (axis = 0) is a parent
    # each column (axis = 1) is a child
    #
    return scipy.sparse.coo_matrix(
        (
            numpy.ones((not_zero_count,)),
            (parent_labels[not_zero], child_labels[not_zero]),
        ),
        shape=(parent_count + 1, child_count + 1),
    )


def relate_histogram(histogram: numpy.ndarray) -> tuple[numpy.ndarray, numpy.ndarray]:
    """Return child counts and parents of children given a histogram

    histogram - histogram from histogram_from_ijv or histogram_from_labels
    """
    parent_count = histogram.shape[0] - 1

    parents_of_children = numpy.asarray(histogram.argmax(axis=0))
    if len(parents_of_children.shape) == 2:
        parents_of_children = numpy.squeeze(parents_of_children, axis=0)
    #
    # Create a histogram of # of children per parent
    children_per_parent = numpy.histogram(
        parents_of_children[1:], numpy.arange(parent_count + 2)
    )[0][1:]

    #
    # Make sure to remove the background elements at index 0
    #
    return children_per_parent, parents_of_children[1:]


def relate_labels(
    parent_labels: numpy.ndarray, child_labels: numpy.ndarray
) -> tuple[numpy.ndarray, numpy.ndarray]:
    """
    relate the object numbers in one label to those in another

    parent_labels - 2d label matrix of parent labels

    child_labels - 2d label matrix of child labels

    Returns two 1-d arrays. The first gives the number of children within
    each parent. The second gives the mapping of each child to its parent's
    object number.

    Ported from https://github.com/cellprofiler/CellProfiler/blob/main/src/subpackages/core/cellprofiler_core/object/_objects.py#L283
    """
    return relate_histogram(find_label_overlaps(parent_labels, child_labels))



import centrosome.cpmorphology
import centrosome.filter
import centrosome.outline
import numpy
import scipy.ndimage
import skimage.segmentation

__doc__ = """
MeasureObjectIntensity
======================

**MeasureObjectIntensity** measures several intensity features for
identified objects.

Given an image with objects identified (e.g., nuclei or cells), this
module extracts intensity features for each object based on one or more
corresponding grayscale images. Measurements are recorded for each
object.

Intensity measurements are made for all combinations of the images and
objects entered. If you want only specific image/object measurements,
you can use multiple MeasureObjectIntensity modules for each group of
measurements desired.

============ ============ ===============
Supports 2D? Supports 3D? Respects masks?
============ ============ ===============
YES          YES          YES
============ ============ ===============

See also
^^^^^^^^

See also **NamesAndTypes**, **MeasureImageIntensity**.

Measurements made by this module
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

-  *IntegratedIntensity:* The sum of the pixel intensities within an
   object.
-  *MeanIntensity:* The average pixel intensity within an object.
-  *StdIntensity:* The standard deviation of the pixel intensities
   within an object.
-  *MaxIntensity:* The maximal pixel intensity within an object.
-  *MinIntensity:* The minimal pixel intensity within an object.
-  *IntegratedIntensityEdge:* The sum of the edge pixel intensities of
   an object.
-  *MeanIntensityEdge:* The average edge pixel intensity of an object.
-  *StdIntensityEdge:* The standard deviation of the edge pixel
   intensities of an object.
-  *MaxIntensityEdge:* The maximal edge pixel intensity of an object.
-  *MinIntensityEdge:* The minimal edge pixel intensity of an object.
-  *MassDisplacement:* The distance between the centers of gravity in
   the gray-level representation of the object and the binary
   representation of the object.
-  *LowerQuartileIntensity:* The intensity value of the pixel for which
   25% of the pixels in the object have lower values.
-  *MedianIntensity:* The median intensity value within the object.
-  *MADIntensity:* The median absolute deviation (MAD) value of the
   intensities within the object. The MAD is defined as the
   median(\|x\ :sub:`i` - median(x)\|).
-  *UpperQuartileIntensity:* The intensity value of the pixel for which
   75% of the pixels in the object have lower values.
-  *Location\_CenterMassIntensity\_X, Location\_CenterMassIntensity\_Y:*
   The (X,Y) coordinates of the intensity weighted centroid (=
   center of mass = first moment) of all pixels within the object.
-  *Location\_MaxIntensity\_X, Location\_MaxIntensity\_Y:* The
   (X,Y) coordinates of the pixel with the maximum intensity within the
   object.
"""

INTENSITY = "Intensity"
INTEGRATED_INTENSITY = "IntegratedIntensity"
MEAN_INTENSITY = "MeanIntensity"
STD_INTENSITY = "StdIntensity"
MIN_INTENSITY = "MinIntensity"
MAX_INTENSITY = "MaxIntensity"
INTEGRATED_INTENSITY_EDGE = "IntegratedIntensityEdge"
MEAN_INTENSITY_EDGE = "MeanIntensityEdge"
STD_INTENSITY_EDGE = "StdIntensityEdge"
MIN_INTENSITY_EDGE = "MinIntensityEdge"
MAX_INTENSITY_EDGE = "MaxIntensityEdge"
MASS_DISPLACEMENT = "MassDisplacement"
LOWER_QUARTILE_INTENSITY = "LowerQuartileIntensity"
MEDIAN_INTENSITY = "MedianIntensity"
MAD_INTENSITY = "MADIntensity"
UPPER_QUARTILE_INTENSITY = "UpperQuartileIntensity"
C_LOCATION = "Location"
LOC_CMI_X = "CenterMassIntensity_X"
LOC_CMI_Y = "CenterMassIntensity_Y"
LOC_CMI_Z = "CenterMassIntensity_Z"
LOC_MAX_X = "MaxIntensity_X"
LOC_MAX_Y = "MaxIntensity_Y"
LOC_MAX_Z = "MaxIntensity_Z"

ALL_MEASUREMENTS = [
    INTEGRATED_INTENSITY,
    MEAN_INTENSITY,
    STD_INTENSITY,
    MIN_INTENSITY,
    MAX_INTENSITY,
    INTEGRATED_INTENSITY_EDGE,
    MEAN_INTENSITY_EDGE,
    STD_INTENSITY_EDGE,
    MIN_INTENSITY_EDGE,
    MAX_INTENSITY_EDGE,
    MASS_DISPLACEMENT,
    LOWER_QUARTILE_INTENSITY,
    MEDIAN_INTENSITY,
    MAD_INTENSITY,
    UPPER_QUARTILE_INTENSITY,
]
ALL_LOCATION_MEASUREMENTS = [
    LOC_CMI_X,
    LOC_CMI_Y,
    LOC_CMI_Z,
    LOC_MAX_X,
    LOC_MAX_Y,
    LOC_MAX_Z,
]


def get_intensity(masks: numpy.ndarray, pixels: numpy.ndarray):
    """
    masks is a labeled array where 0 are background images.
    """
    masked_image = pixels
    image_mask = numpy.ones_like(pixels, dtype=bool)

    if pixels.ndim == 2:
        img = pixels.reshape(1, *pixels.shape)
        masked_image = img

    unique_vals = numpy.unique(masks)
    # MODIFIED: Extract the number of objects by explicitly removing 0s
    nobjects = (unique_vals > 0).sum()

    integrated_intensity = numpy.zeros((nobjects,))
    integrated_intensity_edge = numpy.zeros((nobjects,))
    mean_intensity = numpy.zeros((nobjects,))
    mean_intensity_edge = numpy.zeros((nobjects,))
    std_intensity = numpy.zeros((nobjects,))
    std_intensity_edge = numpy.zeros((nobjects,))
    min_intensity = numpy.zeros((nobjects,))
    min_intensity_edge = numpy.zeros((nobjects,))
    max_intensity = numpy.zeros((nobjects,))
    max_intensity_edge = numpy.zeros((nobjects,))
    mass_displacement = numpy.zeros((nobjects,))
    lower_quartile_intensity = numpy.zeros((nobjects,))
    median_intensity = numpy.zeros((nobjects,))
    mad_intensity = numpy.zeros((nobjects,))
    upper_quartile_intensity = numpy.zeros((nobjects,))
    cmi_x = numpy.zeros((nobjects,))
    cmi_y = numpy.zeros((nobjects,))
    cmi_z = numpy.zeros((nobjects,))
    max_x = numpy.zeros((nobjects,))
    max_y = numpy.zeros((nobjects,))
    max_z = numpy.zeros((nobjects,))

    label_matrices = numpy.zeros((nobjects, *masks.shape), dtype=int)  # N,Y,X
    unique_labels = unique_vals[unique_vals > 0]
    for i, label in enumerate(
        unique_labels
    ):  # Assumes all labels from 1 to nobjects are present
        label_matrices[i][masks == label] = label

    result = {}
    # for labels, lindexes in ((mask, numpy.array([1])),):
    for labels, lindexes in zip(label_matrices, unique_labels):
        lindexes = lindexes[lindexes != 0]

        if pixels.ndim == 2:
            labels = labels.reshape(1, *labels.shape)

        outlines = skimage.segmentation.find_boundaries(labels, mode="inner")

        masked_labels = labels
        masked_outlines = outlines

        lmask = masked_labels > 0 & numpy.isfinite(img)  # Ignore NaNs, Infs
        has_objects = numpy.any(lmask)
        if has_objects:
            limg = img[lmask]

            llabels = labels[lmask]

            mesh_z, mesh_y, mesh_x = numpy.mgrid[
                0 : masked_image.shape[0],
                0 : masked_image.shape[1],
                0 : masked_image.shape[2],
            ]

            mesh_x = mesh_x[lmask]
            mesh_y = mesh_y[lmask]
            mesh_z = mesh_z[lmask]

            lcount = centrosome.cpmorphology.fixup_scipy_ndimage_result(
                scipy.ndimage.sum(numpy.ones(len(limg)), llabels, lindexes)
            )

            integrated_intensity[lindexes - 1] = (
                centrosome.cpmorphology.fixup_scipy_ndimage_result(
                    scipy.ndimage.sum(limg, llabels, lindexes)
                )
            )

            mean_intensity[lindexes - 1] = integrated_intensity[lindexes - 1] / lcount

            std_intensity[lindexes - 1] = numpy.sqrt(
                centrosome.cpmorphology.fixup_scipy_ndimage_result(
                    scipy.ndimage.mean(
                        (limg - mean_intensity[llabels - 1]) ** 2,
                        llabels,
                        lindexes,
                    )
                )
            )

            min_intensity[lindexes - 1] = (
                centrosome.cpmorphology.fixup_scipy_ndimage_result(
                    scipy.ndimage.minimum(limg, llabels, lindexes)
                )
            )

            max_intensity[lindexes - 1] = (
                centrosome.cpmorphology.fixup_scipy_ndimage_result(
                    scipy.ndimage.maximum(limg, llabels, lindexes)
                )
            )

            # Compute the position of the intensity maximum
            max_position = numpy.array(
                centrosome.cpmorphology.fixup_scipy_ndimage_result(
                    scipy.ndimage.maximum_position(limg, llabels, lindexes)
                ),
                dtype=int,
            )
            max_position = numpy.reshape(max_position, (max_position.shape[0],))

            max_x[lindexes - 1] = mesh_x[max_position]
            max_y[lindexes - 1] = mesh_y[max_position]
            max_z[lindexes - 1] = mesh_z[max_position]

            # The mass displacement is the distance between the center
            # of mass of the binary image and of the intensity image. The
            # center of mass is the average X or Y for the binary image
            # and the sum of X or Y * intensity / integrated intensity
            cm_x = centrosome.cpmorphology.fixup_scipy_ndimage_result(
                scipy.ndimage.mean(mesh_x, llabels, lindexes)
            )
            cm_y = centrosome.cpmorphology.fixup_scipy_ndimage_result(
                scipy.ndimage.mean(mesh_y, llabels, lindexes)
            )
            cm_z = centrosome.cpmorphology.fixup_scipy_ndimage_result(
                scipy.ndimage.mean(mesh_z, llabels, lindexes)
            )

            i_x = centrosome.cpmorphology.fixup_scipy_ndimage_result(
                scipy.ndimage.sum(mesh_x * limg, llabels, lindexes)
            )
            i_y = centrosome.cpmorphology.fixup_scipy_ndimage_result(
                scipy.ndimage.sum(mesh_y * limg, llabels, lindexes)
            )
            i_z = centrosome.cpmorphology.fixup_scipy_ndimage_result(
                scipy.ndimage.sum(mesh_z * limg, llabels, lindexes)
            )

            cmi_x[lindexes - 1] = i_x / integrated_intensity[lindexes - 1]
            cmi_y[lindexes - 1] = i_y / integrated_intensity[lindexes - 1]
            cmi_z[lindexes - 1] = i_z / integrated_intensity[lindexes - 1]

            diff_x = cm_x - cmi_x[lindexes - 1]
            diff_y = cm_y - cmi_y[lindexes - 1]
            diff_z = cm_z - cmi_z[lindexes - 1]

            mass_displacement[lindexes - 1] = numpy.sqrt(
                diff_x * diff_x + diff_y * diff_y + diff_z * diff_z
            )

            #
            # Sort the intensities by label, then intensity.
            # For each label, find the index above and below
            # the 25%, 50% and 75% mark and take the weighted
            # average.
            #
            order = numpy.lexsort((limg, llabels))
            areas = lcount.astype(int)
            indices = numpy.cumsum(areas) - areas
            for dest, fraction in (
                (lower_quartile_intensity, 1.0 / 4.0),
                (median_intensity, 1.0 / 2.0),
                (upper_quartile_intensity, 3.0 / 4.0),
            ):
                qindex = indices.astype(float) + areas * fraction
                qfraction = qindex - numpy.floor(qindex)
                qindex = qindex.astype(int)
                qmask = qindex < indices + areas - 1
                qi = qindex[qmask]
                qf = qfraction[qmask]
                dest[lindexes[qmask] - 1] = (
                    limg[order[qi]] * (1 - qf) + limg[order[qi + 1]] * qf
                )

                #
                # In some situations (e.g., only 3 points), there may
                # not be an upper bound.
                #
                qmask = (~qmask) & (areas > 0)
                dest[lindexes[qmask] - 1] = limg[order[qindex[qmask]]]

            #
            # Once again, for the MAD
            #
            madimg = numpy.abs(limg - median_intensity[llabels - 1])
            order = numpy.lexsort((madimg, llabels))
            qindex = indices.astype(float) + areas / pixels.ndim
            qfraction = qindex - numpy.floor(qindex)
            qindex = qindex.astype(int)
            qmask = qindex < indices + areas - 1
            qi = qindex[qmask]
            qf = qfraction[qmask]
            mad_intensity[lindexes[qmask] - 1] = (
                madimg[order[qi]] * (1 - qf) + madimg[order[qi + 1]] * qf
            )
            qmask = (~qmask) & (areas > 0)
            mad_intensity[lindexes[qmask] - 1] = madimg[order[qindex[qmask]]]

            # END OF ITERATIONS

    emask = masked_outlines > 0
    eimg = img[emask]
    elabels = labels[emask]
    has_edge = len(eimg) > 0

    if has_edge:
        ecount = centrosome.cpmorphology.fixup_scipy_ndimage_result(
            scipy.ndimage.sum(numpy.ones(len(eimg)), elabels, lindexes)
        )

        integrated_intensity_edge[lindexes - 1] = (
            centrosome.cpmorphology.fixup_scipy_ndimage_result(
                scipy.ndimage.sum(eimg, elabels, lindexes)
            )
        )

        mean_intensity_edge[lindexes - 1] = (
            integrated_intensity_edge[lindexes - 1] / ecount
        )

        std_intensity_edge[lindexes - 1] = numpy.sqrt(
            centrosome.cpmorphology.fixup_scipy_ndimage_result(
                scipy.ndimage.mean(
                    (eimg - mean_intensity_edge[elabels - 1]) ** 2,
                    elabels,
                )
            )
        )

        min_intensity_edge[lindexes - 1] = (
            centrosome.cpmorphology.fixup_scipy_ndimage_result(
                scipy.ndimage.minimum(eimg, elabels, lindexes)
            )
        )

        max_intensity_edge[lindexes - 1] = (
            centrosome.cpmorphology.fixup_scipy_ndimage_result(
                scipy.ndimage.maximum(eimg, elabels, lindexes)
            )
        )
    for category, feature_name, measurement in (
        (INTENSITY, INTEGRATED_INTENSITY, integrated_intensity),
        (INTENSITY, MEAN_INTENSITY, mean_intensity),
        (INTENSITY, STD_INTENSITY, std_intensity),
        (INTENSITY, MIN_INTENSITY, min_intensity),
        (INTENSITY, MAX_INTENSITY, max_intensity),
        (INTENSITY, INTEGRATED_INTENSITY_EDGE, integrated_intensity_edge),
        (INTENSITY, MEAN_INTENSITY_EDGE, mean_intensity_edge),
        (INTENSITY, STD_INTENSITY_EDGE, std_intensity_edge),
        (INTENSITY, MIN_INTENSITY_EDGE, min_intensity_edge),
        (INTENSITY, MAX_INTENSITY_EDGE, max_intensity_edge),
        (INTENSITY, MASS_DISPLACEMENT, mass_displacement),
        (INTENSITY, LOWER_QUARTILE_INTENSITY, lower_quartile_intensity),
        (INTENSITY, MEDIAN_INTENSITY, median_intensity),
        (INTENSITY, MAD_INTENSITY, mad_intensity),
        (INTENSITY, UPPER_QUARTILE_INTENSITY, upper_quartile_intensity),
        (C_LOCATION, LOC_CMI_X, cmi_x),
        (C_LOCATION, LOC_CMI_Y, cmi_y),
        (C_LOCATION, LOC_CMI_Z, cmi_z),
        (C_LOCATION, LOC_MAX_X, max_x),
        (C_LOCATION, LOC_MAX_Y, max_y),
        (C_LOCATION, LOC_MAX_Z, max_z),
    ):
        measurement_name = "{}_{}".format(category, feature_name)
        # MODIFIED: Selected first
        result[measurement_name] = measurement

    return result



import centrosome.cpmorphology
import centrosome.propagate
import centrosome.zernike
import numpy
import numpy.ma
import scipy.ndimage
import scipy.sparse

from cp_measure.utils import masks_to_ijv

""""
============ ============ ===============
Supports 2D? Supports 3D? Respects masks?
============ ============ ===============
YES          NO           YES
============ ============ ===============

See also
^^^^^^^^

See also **MeasureObjectIntensity** and **/MeasureTexture**.

Measurements made by this module
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

-  *FracAtD:* Fraction of total stain in an object at a given radius.
-  *MeanFrac:* Mean fractional intensity at a given radius; calculated
   as fraction of total intensity normalized by fraction of pixels at a
   given radius.
-  *RadialCV:* Coefficient of variation of intensity within a ring,
   calculated across 8 slices.
-  *Zernike:* The Zernike features characterize the distribution of
   intensity across the object. For instance, Zernike 1,1 has a high
   value if the intensity is low on one side of the object and high on
   the other. The ZernikeMagnitude feature records the rotationally
   invariant degree magnitude of the moment and the ZernikePhase feature
   gives the moment’s orientation.

"""

Z_NONE = "None"
Z_MAGNITUDES = "Magnitudes only"
Z_MAGNITUDES_AND_PHASE = "Magnitudes and phase"
Z_ALL = [Z_NONE, Z_MAGNITUDES, Z_MAGNITUDES_AND_PHASE]

M_CATEGORY = "RadialDistribution"
F_FRAC_AT_D = "FracAtD"
F_MEAN_FRAC = "MeanFrac"
F_RADIAL_CV = "RadialCV"
F_ALL = [F_FRAC_AT_D, F_MEAN_FRAC, F_RADIAL_CV]

FF_SCALE = "%dof%d"
FF_OVERFLOW = "Overflow"
FF_GENERIC = FF_SCALE

MF_FRAC_AT_D = "_".join((M_CATEGORY, F_FRAC_AT_D, FF_GENERIC))
MF_MEAN_FRAC = "_".join((M_CATEGORY, F_MEAN_FRAC, FF_GENERIC))
MF_RADIAL_CV = "_".join((M_CATEGORY, F_RADIAL_CV, FF_GENERIC))
OF_FRAC_AT_D = "_".join((M_CATEGORY, F_FRAC_AT_D, FF_OVERFLOW))
OF_MEAN_FRAC = "_".join((M_CATEGORY, F_MEAN_FRAC, FF_OVERFLOW))
OF_RADIAL_CV = "_".join((M_CATEGORY, F_RADIAL_CV, FF_OVERFLOW))

"""# of settings aside from groups"""
SETTINGS_STATIC_COUNT = 3
"""# of settings in image group"""
SETTINGS_IMAGE_GROUP_COUNT = 1
"""# of settings in object group"""
SETTINGS_OBJECT_GROUP_COUNT = 3
"""# of settings in bin group, v1"""
SETTINGS_BIN_GROUP_COUNT_V1 = 1
"""# of settings in bin group, v2"""
SETTINGS_BIN_GROUP_COUNT_V2 = 3
SETTINGS_BIN_GROUP_COUNT = 3
"""# of settings in heatmap group, v4"""
SETTINGS_HEATMAP_GROUP_COUNT_V4 = 7
SETTINGS_HEATMAP_GROUP_COUNT = 7
"""Offset of center choice in object group"""
SETTINGS_CENTER_CHOICE_OFFSET = 1

A_FRAC_AT_D = "Fraction at Distance"
A_MEAN_FRAC = "Mean Fraction"
A_RADIAL_CV = "Radial CV"
MEASUREMENT_CHOICES = [A_FRAC_AT_D, A_MEAN_FRAC, A_RADIAL_CV]

MEASUREMENT_ALIASES = {
    A_FRAC_AT_D: MF_FRAC_AT_D,
    A_MEAN_FRAC: MF_MEAN_FRAC,
    A_RADIAL_CV: MF_RADIAL_CV,
}


def get_radial_distribution(
    labels: numpy.ndarray,
    pixels: numpy.ndarray,
    scaled: bool = True,
    bin_count: int = 4,
    maximum_radius: int = 100,
):
    """
    zernike_degree : int
        Maximum zernike moment.

        This is the maximum radial moment that will be calculated. There are
        increasing numbers of azimuthal moments as you increase the radial
        moment, so higher values are increasingly expensive to calculate.

    scaled : bool
        Scale the bins?

        When True divide the object radially into the number of bins
        that you specify. Otherwise create the number of bins you specify
        based on distance. If True, it will use a maximum distance so
        that each object will have the same measurements (which might be zero
        for small objects) and so that the measurements can be taken without
        knowing the maximum object radius before the run starts.

    bin_count : int
        Number of bins

        Specify the number of bins that you want to use to measure the distribution.
        Radial distribution is measured with respect to a series of concentric
        rings starting from the object center (or more generally, between contours
        at a normalized distance from the object center). This number specifies
        the number of rings into which the distribution is to be divided.

    maximum_radius : int
        Maximum radius

        Specify the maximum radius for the unscaled bins. The unscaled binning method
        creates the number of bins that you specify and creates equally spaced bin
        boundaries up to the maximum radius. Parts of the object that are beyond this
        radius will be counted in an overflow bin. The radius is measured in pixels.
    """

    if labels.dtype == bool:
        labels = labels.astype(numpy.integer)

    unique_labels = numpy.unique(labels)
    nobjects = len(unique_labels[unique_labels > 0])
    d_to_edge = centrosome.cpmorphology.distance_to_edge(labels)

    # Find the point in each object farthest away from the edge.
    # This does better than the centroid:
    # * The center is within the object
    # * The center tends to be an interesting point, like the
    #   center of the nucleus or the center of one or the other
    #   of two touching cells.
    #
    # MODIFICATION: Delegated label indices to maximum_position_of_labels
    # This should not affect this one-mask/object function
    i, j = centrosome.cpmorphology.maximum_position_of_labels(
        # d_to_edge, labels, indices=[1]
        d_to_edge,
        labels,
        indices=[1],
    )

    center_labels = numpy.zeros(labels.shape, int)

    center_labels[i, j] = labels[i, j]

    #
    # Use the coloring trick here to process touching objects
    # in separate operations
    #
    colors = centrosome.cpmorphology.color_labels(labels)

    ncolors = numpy.max(colors)

    d_from_center = numpy.zeros(labels.shape)

    cl = numpy.zeros(labels.shape, int)

    for color in range(1, ncolors + 1):
        mask = colors == color
        l, d = centrosome.propagate.propagate(
            numpy.zeros(center_labels.shape), center_labels, mask, 1
        )

        d_from_center[mask] = d[mask]

        cl[mask] = l[mask]

    good_mask = cl > 0

    i_center = numpy.zeros(cl.shape)

    i_center[good_mask] = i[cl[good_mask] - 1]

    j_center = numpy.zeros(cl.shape)

    j_center[good_mask] = j[cl[good_mask] - 1]

    normalized_distance = numpy.zeros(labels.shape)

    if scaled:
        total_distance = d_from_center + d_to_edge

        normalized_distance[good_mask] = d_from_center[good_mask] / (
            total_distance[good_mask] + 0.001
        )
    else:
        normalized_distance[good_mask] = d_from_center[good_mask] / maximum_radius

    ngood_pixels = numpy.sum(good_mask)

    good_labels = labels[good_mask]

    bin_indexes = (normalized_distance * bin_count).astype(int)

    bin_indexes[bin_indexes > bin_count] = bin_count

    labels_and_bins = (good_labels - 1, bin_indexes[good_mask])

    histogram = scipy.sparse.coo_matrix(
        (pixels[good_mask], labels_and_bins), (nobjects, bin_count + 1)
    ).toarray()

    sum_by_object = numpy.sum(histogram, 1)

    sum_by_object_per_bin = numpy.dstack([sum_by_object] * (bin_count + 1))[0]

    fraction_at_distance = histogram / sum_by_object_per_bin

    number_at_distance = scipy.sparse.coo_matrix(
        (numpy.ones(ngood_pixels), labels_and_bins), (nobjects, bin_count + 1)
    ).toarray()

    sum_by_object = numpy.sum(number_at_distance, 1)

    sum_by_object_per_bin = numpy.dstack([sum_by_object] * (bin_count + 1))[0]

    fraction_at_bin = number_at_distance / sum_by_object_per_bin

    mean_pixel_fraction = fraction_at_distance / (
        fraction_at_bin + numpy.finfo(float).eps
    )

    # Anisotropy calculation.  Split each cell into eight wedges, then
    # compute coefficient of variation of the wedges' mean intensities
    # in each ring.
    #
    # Compute each pixel's delta from the center object's centroid
    i, j = numpy.mgrid[0 : labels.shape[0], 0 : labels.shape[1]]

    imask = i[good_mask] > i_center[good_mask]

    jmask = j[good_mask] > j_center[good_mask]

    absmask = abs(i[good_mask] - i_center[good_mask]) > abs(
        j[good_mask] - j_center[good_mask]
    )

    radial_index = imask.astype(int) + jmask.astype(int) * 2 + absmask.astype(int) * 4

    results = {}

    for bin in range(bin_count + (0 if scaled else 1)):
        bin_mask = good_mask & (bin_indexes == bin)

        bin_pixels = numpy.sum(bin_mask)

        bin_labels = labels[bin_mask]

        bin_radial_index = radial_index[bin_indexes[good_mask] == bin]

        labels_and_radii = (bin_labels - 1, bin_radial_index)

        radial_values = scipy.sparse.coo_matrix(
            (pixels[bin_mask], labels_and_radii), (nobjects, 8)
        ).toarray()

        pixel_count = scipy.sparse.coo_matrix(
            (numpy.ones(bin_pixels), labels_and_radii), (nobjects, 8)
        ).toarray()

        mask = pixel_count == 0

        radial_means = numpy.ma.masked_array(radial_values / pixel_count, mask)

        radial_cv = numpy.std(radial_means, 1) / numpy.mean(radial_means, 1)

        radial_cv[numpy.sum(~mask, 1) == 0] = 0

        for measurement, feature, overflow_feature in (
            (fraction_at_distance[:, bin], MF_FRAC_AT_D, OF_FRAC_AT_D),
            (mean_pixel_fraction[:, bin], MF_MEAN_FRAC, OF_MEAN_FRAC),
            (numpy.array(radial_cv), MF_RADIAL_CV, OF_RADIAL_CV),
        ):
            if bin == bin_count:
                measurement_name = overflow_feature
            else:
                measurement_name = feature % (bin + 1, bin_count)

            results[measurement_name] = measurement

    return results


def get_radial_zernikes(
    labels: numpy.ndarray, pixels: numpy.ndarray, zernike_degree: int = 9
):
    zernike_indexes = centrosome.zernike.get_zernike_indexes(zernike_degree + 1)

    unique_labels = numpy.unique(labels)  # Will be used later for scipy.ndimage.sum
    unique_labels = unique_labels[unique_labels > 0]
    # MODIFIED: Delegate index generation to the minimum_enclosing_circle
    # TODO: Check that this has the correct dimensions
    ij = numpy.zeros((len(unique_labels) + 1, 2))
    r = numpy.zeros((len(unique_labels) + 1))
    # MODIFIED: We assume non-overlapping labels for now
    # TODO Support label overlap (i.e., format in ijv)
    # MODIFIED: Delegate indexes to minimum_enclosing_circle
    for index, _ in enumerate(unique_labels, 1):
        ij_, r_ = centrosome.cpmorphology.minimum_enclosing_circle(labels, [index])
        ij[[index]] = ij_
        r[[index]] = r_

    #
    # Then compute x and y, the position of each labeled pixel
    # within a unit circle around the object
    #
    ijv = masks_to_ijv(labels)

    # MODIFIED: TODO double-check check that this -1 makes sense! `l` is used as indices later on
    l = ijv[:, 2]

    yx = (ijv[:, :2] - ij[l, :]) / r[l, numpy.newaxis]

    z = centrosome.zernike.construct_zernike_polynomials(
        yx[:, 1], yx[:, 0], zernike_indexes
    )

    # Filter ijv-formatted items to keep the ones inside the pixels boundary
    ijv_mask = (ijv[:, 0] < pixels.shape[0]) & (ijv[:, 1] < pixels.shape[1])
    # ijv_mask[ijv_mask] = pixels[ijv[ijv_mask,0], ijv[ijv_mask, 1]]

    yx = yx[ijv_mask, :]
    l_ = l[ijv_mask]
    z_ = z[ijv_mask, :]

    if len(l_) == 0:
        # Cover fringe case in which all labels were filtered out
        results = {}
        for mag_or_phase in ("Magnitude", "Phase"):
            for n, m in zernike_indexes:
                name = f"{M_CATEGORY}_Zernike{mag_or_phase}_{n}_{m}"
                results[name] = numpy.zeros(0)
    else:
        # MODIFIED: Replaced sum with the updated sum_labels
        areas = scipy.ndimage.sum_labels(
            numpy.ones(l_.shape, int), labels=l_, index=unique_labels
        )

        #
        # Results will be formatted in a dictionary with the following keys:
        # Zernike{Magniture|Phase}_{n}_{m}
        # n - the radial moment of the Zernike
        # m - the azimuthal moment of the Zernike
        #
        results = {}
        for i, (n, m) in enumerate(zernike_indexes):
            vr = scipy.ndimage.sum_labels(
                pixels[ijv[:, 0], ijv[:, 1]] * z_[:, i].real,
                labels=l_,
                index=unique_labels,
            )

            vi = scipy.ndimage.sum_labels(
                pixels[ijv[:, 0], ijv[:, 1]] * z[:, i].imag,
                labels=l_,
                index=unique_labels,
            )

            magnitude = numpy.sqrt(vr * vr + vi * vi) / areas
            phase = numpy.arctan2(vr, vi)

            results[f"{M_CATEGORY}_ZernikeMagnitude_{n}_{m}"] = magnitude
            results[f"{M_CATEGORY}_ZernikePhase_{n}_{m}"] = phase

    return results



""" "
Utilities reused in multiple measurements.
"""

import numpy


def get_test_pixels_mask():
    pixels = numpy.random.randint(100, size=64**2).reshape((64, 64))
    mask = numpy.zeros_like(pixels, dtype=bool)
    mask[2:-3, 2:-3] = True
    return pixels, mask


def masks_to_ijv(masks: numpy.ndarray) -> numpy.ndarray:
    """
    input: 2d boolean array
    output: (n, 3) integer array following (i,j,1)
    """

    # Extract coordinates of object from boolean mask
    masks_ijv = numpy.empty((0, 3), dtype=int)
    for label in range(masks.max()):
        i, j = numpy.where(masks == label + 1)
        n = len(i)
        ijv = numpy.empty((n, 3), dtype=int)
        ijv[:, 0] = i
        ijv[:, 1] = j
        ijv[:, 2] = label + 1
        masks_ijv = numpy.concatenate((masks_ijv, ijv))

    return masks_ijv


def labels_to_binmasks(masks: numpy.ndarray) -> numpy.ndarray:
    """
    Convert a label matrix to a boolean masks.

    Returns a list of binary masks.
    """
    labels = numpy.unique(masks)
    labels = sorted(labels[labels > 0])
    return [(masks == i) for i in labels]



__doc__ = """\
MeasureTexture
==============

**MeasureTexture** measures the degree and nature of textures within
images and objects to quantify their roughness and smoothness.

This module measures intensity variations in grayscale images. An object or
entire image without much texture has a smooth appearance; an object or
image with a lot of texture will appear rough and show a wide variety of
pixel intensities.

Note that any input objects specified will have their texture measured
against *all* input images specified, which may lead to image-object
texture combinations that are unnecessary. If you do not want this
behavior, use multiple **MeasureTexture** modules to specify the
particular image-object measures that you want.

Note also that CellProfiler in all 2.X versions increased speed by binning 
the image into only 8 grayscale levels before calculating Haralick features;
in all 3.X CellProfiler versions the images were binned into 256 grayscale
levels. CellProfiler 4 allows you to select your own preferred number of
grayscale levels, but note that since we use a slightly different
implementation than CellProfiler 2 we do not guarantee concordance with
CellProfiler 2.X-generated texture values.

|

============ ============ ===============
Supports 2D? Supports 3D? Respects masks?
============ ============ ===============
YES          YES          YES
============ ============ ===============

Measurements made by this module
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

-  *Haralick Features:* Haralick texture features are derived from the
   co-occurrence matrix, which contains information about how image
   intensities in pixels with a certain position in relation to each
   other occur together. **MeasureTexture** can measure textures at
   different scales; the scale you choose determines how the
   co-occurrence matrix is constructed. For example, if you choose a
   scale of 2, each pixel in the image (excluding some border pixels)
   will be compared against the one that is two pixels to the right.

   Thirteen measurements are then calculated for the image by performing
   mathematical operations on the co-occurrence matrix (the formulas can
   be found `here`_):

   -  *AngularSecondMoment:* Measure of image homogeneity. A higher
      value of this feature indicates that the intensity varies less in
      an image. Has a value of 1 for a uniform image.
   -  *Contrast:* Measure of local variation in an image, with 0 for a
      uniform image and a high value indicating a high degree of local
      variation.
   -  *Correlation:* Measure of linear dependency of intensity values in
      an image. For an image with large areas of similar intensities,
      correlation is much higher than for an image with noisier,
      uncorrelated intensities. Has a value of 1 or -1 for a perfectly
      positively or negatively correlated image, respectively.
   -  *Variance:* Measure of the variation of image intensity values.
      For an image with uniform intensity, the texture variance would be
      zero.
   -  *InverseDifferenceMoment:* Another feature to represent image
      contrast. Has a low value for inhomogeneous images, and a
      relatively higher value for homogeneous images.
   -  *SumAverage:* The average of the normalized grayscale image in the
      spatial domain.
   -  *SumVariance:* The variance of the normalized grayscale image in
      the spatial domain.
   -  *SumEntropy:* A measure of randomness within an image.
   -  *Entropy:* An indication of the complexity within an image. A
      complex image produces a high entropy value.
   -  *DifferenceVariance:* The image variation in a normalized
      co-occurrence matrix.
   -  *DifferenceEntropy:* Another indication of the amount of
      randomness in an image.
   -  *InfoMeas1:* A measure of the total amount of information contained
      within a region of pixels derived from the recurring spatial
      relationship between specific intensity values.
   -  *InfoMeas2:* An additional measure of the total amount of information
      contained within a region of pixels derived from the recurring spatial
      relationship between specific intensity values. It is a complementary
      value to InfoMeas1 and is on a different scale.

**Note**: each of the above measurements are computed for different 
'directions' in the image, specified by a series of correspondence vectors. 
These are indicated in the results table in the *scale* column as n_00, n_01,
n_02... for each scale *n*. In 2D, the directions and correspondence vectors *(y, x)* 
for each measurement are given below:

- _00 = horizontal -, 0 degrees   (0, 1)
- _01 = diagonal \\\\, 135 degrees or NW-SE   (1, 1)
- _02 = vertical \|, 90 degrees   (1, 0)
- _03 = diagonal /, 45 degrees or NE-SW  (1, -1)

When analyzing 3D images, there are 13 correspondence vectors *(y, x, z)*:

- (1, 0, 0)
- (1, 1, 0)
- (0, 1, 0)
- (1,-1, 0)
- (0, 0, 1)
- (1, 0, 1)
- (0, 1, 1)
- (1, 1, 1)
- (1,-1, 1)
- (1, 0,-1)
- (0, 1,-1)
- (1, 1,-1)
- (1,-1,-1)

In this case, an image makes understanding their directions easier. 
Imagine the origin (0, 0, 0) is at the upper left corner of the first image
in your z-stack. Yellow vectors fall along the axes, and pairs of vectors with 
matching colors are reflections of each other across the x axis. The two
images represent two views of the same vectors. Images made in `GeoGebra`_.

|MT_image0| |MT_image1|

Technical notes
^^^^^^^^^^^^^^^

To calculate the Haralick features, **MeasureTexture** normalizes the
co-occurrence matrix at the per-object level by basing the intensity
levels of the matrix on the maximum and minimum intensity observed
within each object. This is beneficial for images in which the maximum
intensities of the objects vary substantially because each object will
have the full complement of levels.

References
^^^^^^^^^^

-  Haralick RM, Shanmugam K, Dinstein I. (1973), “Textural Features for
   Image Classification” *IEEE Transaction on Systems Man, Cybernetics*,
   SMC-3(6):610-621. `(link) <https://doi.org/10.1109/TSMC.1973.4309314>`__

.. _here: http://murphylab.web.cmu.edu/publications/boland/boland_node26.html
.. _GeoGebra: https://www.geogebra.org/ 
.. |MT_image0| image:: {MEASURE_TEXTURE_3D_INFO}
.. |MT_image1| image:: {MEASURE_TEXTURE_3D_INFO2}
"""

import mahotas.features
import numpy
import skimage.exposure
import skimage.measure
import skimage.util

F_HARALICK = """AngularSecondMoment Contrast Correlation Variance
InverseDifferenceMoment SumAverage SumVariance SumEntropy Entropy
DifferenceVariance DifferenceEntropy InfoMeas1 InfoMeas2""".split()


def get_texture(
    masks: numpy.ndarray,
    pixels: numpy.ndarray,
    scale: int = 3,
    gray_levels: int = 256,
):
    """
    Parameters
    ----------
    gray_levels : int, optional (default is 256)
        Number of gray levels. Measuring at more levels gives you _potentially_
        more detailed information about your image, but at the cost of somewhat
        decreased processing speed (default is 256).
    texture_scale : int, optional (default is 3)
        You can specify the scale of texture to be measured, in pixel units; the
        texture scale is the distance between correlated intensities in the
        image. A higher number for the scale of texture measures larger patterns
        of texture whereas smaller numbers measure more localized patterns of
        texture. It is best to measure texture on a scale smaller than your
        objects’ sizes, so be sure that the value entered for scale of texture
        is smaller than most of your objects. For very small objects (smaller
        than the scale of texture you are measuring), the texture cannot be
        measured and will result in a undefined value in the output file.

    Returns
    -------
    Dictionary of list of floats.

    Notes
    -----
    Before processing, your image will be rescaled from its current pixel values
    to 0 - [gray levels - 1]. The texture features will then be calculated.

    In all CellProfiler 2 versions, this value was fixed at 8; in all
    CellProfiler 3 versions it was fixed at 256.  The minimum number of levels is
    2, the maximum is 256.
    """
    unique_labels = numpy.unique(masks)
    unique_labels = unique_labels[unique_labels > 0]

    # Modified to use the number of dimensions in pixels to determine the number of directions
    n_directions = 13 if pixels.ndim > 2 else 4

    # MODIFIED: We assume that the mask provided has the same shape
    # as pixels, thus no cropping is performed
    pixels[~masks.astype(bool)] = 0

    # mahotas.features.haralick bricks itself when provided a
    # dtype larger than uint8 (version 1.4.3)
    pixels = skimage.util.img_as_ubyte(pixels)
    if gray_levels != 256:
        pixels = skimage.exposure.rescale_intensity(
            pixels, in_range=(0, 255), out_range=(0, gray_levels - 1)
        ).astype(numpy.uint8)
    props = skimage.measure.regionprops(masks, pixels)

    features = numpy.empty((n_directions, 13, len(unique_labels)))

    for index, prop in enumerate(props):
        label_data = prop["intensity_image"]
        try:
            features[:, :, index] = mahotas.features.haralick(
                label_data, distance=scale, ignore_zeros=True
            )
        except ValueError:
            features[:, :, index] = numpy.nan

    # MODIFIED: Reconstructed name:
    # Texture_{X}_{scale}_{direction_id}_{graylevels}
    results = {}
    for direction_i, direction_features in enumerate(features):
        for feature_name, values in zip(F_HARALICK, direction_features):
            results[
                "{}_{:d}_{:02d}_{:d}".format(
                    feature_name,
                    scale,
                    direction_i,
                    gray_levels,
                )
            ] = values

    return results



"""
MeasureObjectOverlap
====================

**MeasureObjectOverlap** calculates how much overlap occurs between
objects.

This module calculates object overlap by determining a set of statistics
that measure the closeness of an object to its true value. One
object is considered the “ground truth” (possibly the result of
hand-segmentation) and the other is the “test” object; the objects
are determined to overlap most completely when the test object matches
the ground truth perfectly. The module requires input to be objects obtained
after "IdentifyPrimaryObjects", "IdentifySecondaryObjects" or "IdentifyTertiaryObjects".
If your images have been segmented using other image processing software,
or you have hand-segmented them in software such as Photoshop, you will
need to use "Object Processing" modules such as "IdentifyPrimaryObjects" to identify
"ground truth" objects.

Measurements made by this module
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

-  *True positive rate:* Total number of true positive pixels / total number of actual positive pixels.

-  *False positive rate:* Total number of false positive pixels / total number of actual negative pixels

-  *True negative rate:* Total number of true negative pixels / total number of actual negative pixels.

-  *False negative rate:* Total number of false negative pixels / total number of actual positive pixels

-  *Precision:* Number of true positive pixels / (number of true positive pixels + number of false positive pixels)

-  *Recall:* Number of true positive pixels/ (number of true positive pixels + number of false negative pixels)

-  *F-factor:* 2 × (precision × recall)/(precision + recall). Also known as F\ :sub:`1` score, F-score or F-measure.

-  *Earth mover’s distance:* The minimum distance required to move each foreground
   pixel in the test object to some corresponding foreground pixel in the reference object.

-  *Rand index:* A measure of the similarity between two data clusterings. Perfectly random clustering
   returns the minimum score of 0, perfect clustering returns the maximum score of 1.

-  *Adjusted Rand index:* A variation of the Rand index which considers a correction for chance.

References
^^^^^^^^^^

-  Collins LM, Dent CW (1988) “Omega: A general formulation of the Rand
   Index of cluster recovery suitable for non-disjoint solutions”,
   *Multivariate Behavioral Research*, 23, 231-242 `(link)`_

-  Pele O, Werman M (2009) “Fast and Robust Earth Mover’s Distances”,
   *2009 IEEE 12th International Conference on Computer Vision*

.. _(link): https://doi.org/10.1207/s15327906mbr2302_6
"""

from functools import reduce

import centrosome.cpmorphology
import centrosome.fastemd
import centrosome.filter
import centrosome.index
import centrosome.propagate
import numpy
import scipy.ndimage
import scipy.sparse
from cp_measure.utils import labels_to_binmasks, masks_to_ijv

C_IMAGE_OVERLAP = "Overlap"
FTR_F_FACTOR = "Ffactor"
FTR_PRECISION = "Precision"
FTR_RECALL = "Recall"
FTR_TRUE_POS_RATE = "TruePosRate"
FTR_FALSE_POS_RATE = "FalsePosRate"
FTR_FALSE_NEG_RATE = "FalseNegRate"
FTR_TRUE_NEG_RATE = "TrueNegRate"
FTR_RAND_INDEX = "RandIndex"
FTR_ADJUSTED_RAND_INDEX = "AdjustedRandIndex"
FTR_EARTH_MOVERS_DISTANCE = "EarthMoversDistance"

FTR_ALL = [
    FTR_F_FACTOR,
    FTR_PRECISION,
    FTR_RECALL,
    FTR_TRUE_POS_RATE,
    FTR_TRUE_NEG_RATE,
    FTR_FALSE_POS_RATE,
    FTR_FALSE_NEG_RATE,
    FTR_RAND_INDEX,
    FTR_ADJUSTED_RAND_INDEX,
]

O_OBJ = "Segmented objects"

L_LOAD = "Loaded from a previous run"
L_CP = "From this CP pipeline"

DM_KMEANS = "K Means"
DM_SKEL = "Skeleton"


def nan_divide(
    numerator: numpy.float64 | float | int, denominator: numpy.float64 | float | int
) -> float:
    if denominator == 0:
        return numpy.nan
    return float(numerator) / float(denominator)


def subscripts(
    condition1: numpy.ndarray,
    condition2: numpy.ndarray,
    GT_pixels: numpy.ndarray,
    ID_pixels: numpy.ndarray,
) -> list:
    x1, y1 = numpy.where(GT_pixels == condition1)
    x2, y2 = numpy.where(ID_pixels == condition2)
    mask = set(zip(x1, y1)) & set(zip(x2, y2))
    return list(mask)


def maskimg(mask, img):
    for ea in mask:
        img[ea] = 1
    return img


def measureobjectoverlap(
    masks1: numpy.ndarray,
    masks2: numpy.ndarray,
    wants_emd: bool = False,
    max_points: int = 250,
    decimation_method: str = DM_KMEANS,
    max_distance: int = 100,
    penalize_missing: bool = False,
) -> dict[str, float]:
    """
    Assumes 2D masks shown as 
    wants_emd:
        Calculate earth mover's distance?
        The earth mover’s distance computes the shortest distance that would
        have to be travelled to move each foreground pixel in the test object to
        some foreground pixel in the reference object. “Earth mover’s” refers to
        an analogy: the pixels are “earth” that has to be moved by some machine
        at the smallest possible cost.
        It would take too much memory and processing time to compute the exact
        earth mover’s distance, so **MeasureObjectOverlap** chooses
        representative foreground pixels in each object and assigns each
        foreground pixel to its closest representative. The earth mover’s
        distance is then computed for moving the foreground pixels associated
        with each representative in the test object to those in the reference
        object.
    
    max_points:
        *(Used only when computing the earth mover’s distance)*

        This is the number of representative points that will be taken from the
        foreground of the test objects and from the foreground of the reference
        objects using the point selection method (see below). Minimum value (100).

    decimation_method:
        Point selection method, *(Used only when computing the earth mover’s distance)*

        Options: {DM_KMEANS}, {DM_SKEL}
        The point selection setting determines how the representative points
        are chosen.

        -  *{DM_KMEANS}:* Select to pick representative points using a K-Means
           clustering technique. The foregrounds of both objects are combined and
           representatives are picked that minimize the distance to the nearest
           representative. The same representatives are then used for the test
           and reference objects.
        -  *{DM_SKEL}:* Select to skeletonize the object and pick points
           equidistant along the skeleton.
    
        *{DM_KMEANS}* is a choice that’s generally applicable to all
        images. *{DM_SKEL}* is best suited to long, skinny objects such as
        worms or neurites.
    
    max_distance: int
        Maximum distance, *(Used only when computing the earth mover’s distance)*

        This setting sets an upper bound to the distance penalty assessed during
        the movement calculation. As an example, the score for moving 10 pixels
        from one location to a location that is 100 pixels away is 10\*100, but
        if the maximum distance were set to 50, the score would be 10\*50
        instead.

        The maximum distance should be set to the largest reasonable distance
        that pixels could be expected to move from one object to the next. Minimum value is 1.

    penalize_missing: bool
        Penalize missing pixels, *(Used only when computing the earth mover’s distance)*

        If one object has more foreground pixels than the other, the earth
        mover’s distance is not well-defined because there is no destination for
        the extra source pixels or vice-versa. It’s reasonable to assess a
        penalty for the discrepancy when comparing the accuracy of a
        segmentation because the discrepancy represents an error. It’s also
        reasonable to assess no penalty if the goal is to compute the cost of
        movement, for example between two frames in a time-lapse movie, because
        the discrepancy is likely caused by noise or artifacts in segmentation.
        Set this setting to “Yes” to assess a penalty equal to the maximum
        distance times the absolute difference in number of foreground pixels in
        the two objects. Set this setting to “No” to assess no penalty.
    """.format(
        **{
            "DM_KMEANS": DM_KMEANS,
            "DM_SKEL": DM_SKEL,
        }
    )
    # category = "Measurement"
    # variable_revision_number = 2
    # module_name = "MeasureObjectOverlap"

    # object_name_GT = self.object_name_GT.value
    # objects_GT = workspace.get_objects(object_name_GT)
    objects_GT = masks1
    # iGT, jGT, lGT = objects_GT.ijv.transpose()
    iGT, jGT, lGT = masks_to_ijv(objects_GT).transpose()
    # object_name_ID = self.object_name_ID.value
    # objects_ID = workspace.get_objects(object_name_ID)
    objects_ID = masks2
    iID, jID, lID = masks_to_ijv(objects_ID).transpose()
    ID_obj = 0 if len(lID) == 0 else max(lID)
    GT_obj = 0 if len(lGT) == 0 else max(lGT)

    xGT, yGT = objects_GT.shape
    xID, yID = objects_ID.shape
    GT_pixels = numpy.zeros((xGT, yGT))
    ID_pixels = numpy.zeros((xID, yID))
    total_pixels = xGT * yGT

    GT_pixels[iGT, jGT] = 1
    ID_pixels[iID, jID] = 1

    GT_tot_area = len(iGT)
    if len(iGT) == 0 and len(iID) == 0:
        intersect_matrix = numpy.zeros((0, 0), int)
    else:
        #
        # Build a matrix with rows of i, j, label and a GT/ID flag
        #
        all_ijv = numpy.column_stack(
            (
                numpy.hstack((iGT, iID)),
                numpy.hstack((jGT, jID)),
                numpy.hstack((lGT, lID)),
                numpy.hstack((numpy.zeros(len(iGT)), numpy.ones(len(iID)))),
            )
        )
        #
        # Order it so that runs of the same i, j are consecutive
        #
        order = numpy.lexsort((all_ijv[:, -1], all_ijv[:, 0], all_ijv[:, 1]))
        all_ijv = all_ijv[order, :]
        # Mark the first at each i, j != previous i, j
        first = numpy.where(
            numpy.hstack(
                ([True], ~numpy.all(all_ijv[:-1, :2] == all_ijv[1:, :2], 1), [True])
            )
        )[0]
        # Count # at each i, j
        count = first[1:] - first[:-1]
        # First indexer - mapping from i,j to index in all_ijv
        all_ijv_map = centrosome.index.Indexes([count])
        # Bincount to get the # of ID pixels per i,j
        id_count = numpy.bincount(all_ijv_map.rev_idx, all_ijv[:, -1]).astype(int)
        gt_count = count - id_count
        # Now we can create an indexer that has NxM elements per i,j
        # where N is the number of GT pixels at that i,j and M is
        # the number of ID pixels. We can then use the indexer to pull
        # out the label values for each to populate a sparse array.
        #
        cross_map = centrosome.index.Indexes([id_count, gt_count])
        off_gt = all_ijv_map.fwd_idx[cross_map.rev_idx] + cross_map.idx[0]
        off_id = (
            all_ijv_map.fwd_idx[cross_map.rev_idx]
            + cross_map.idx[1]
            + id_count[cross_map.rev_idx]
        )
        intersect_matrix = scipy.sparse.coo_matrix(
            (numpy.ones(len(off_gt)), (all_ijv[off_id, 2], all_ijv[off_gt, 2])),
            shape=(ID_obj + 1, GT_obj + 1),
        ).toarray()[1:, 1:]

        # TODO adjust
        gt_areas = get_areas(objects_GT)
        id_areas = get_areas(objects_ID)
        FN_area = gt_areas[numpy.newaxis, :] - intersect_matrix
        # all_intersecting_area = numpy.sum(intersect_matrix)
        # all_intersecting_area = numpy.sum(intersect_matrix)
        dom_ID = []

        for i in range(0, ID_obj):
            indices_jj = numpy.nonzero(lID == i)
            indices_jj = indices_jj[0]
            id_i = iID[indices_jj]
            id_j = jID[indices_jj]
            ID_pixels[id_i, id_j] = 1

        for i in intersect_matrix:  # loop through the GT objects first
            if len(i) == 0 or max(i) == 0:
                id = -1  # we missed the object; arbitrarily assign -1 index
            else:
                id = numpy.where(i == max(i))[0][0]  # what is the ID of the max pixels?
            dom_ID += [id]  # for ea GT object, which is the dominating ID?

        dom_ID = numpy.array(dom_ID)

        for i in range(0, len(intersect_matrix.T)):
            if len(numpy.where(dom_ID == i)[0]) > 1:
                final_id = numpy.where(
                    intersect_matrix.T[i] == max(intersect_matrix.T[i])
                )
                final_id = final_id[0][0]
                all_id = numpy.where(dom_ID == i)[0]
                nonfinal = [x for x in all_id if x != final_id]
                for (
                    n
                ) in nonfinal:  # these others cannot be candidates for the corr ID now
                    intersect_matrix.T[i][n] = 0
            else:
                continue

        TP = 0
        FN = 0
        FP = 0
        for i in range(0, len(dom_ID)):
            d = dom_ID[i]
            if d == -1:
                tp = 0
                fn = id_areas[i]
                fp = 0
            else:
                fp = numpy.sum(intersect_matrix[i][0:d]) + numpy.sum(
                    intersect_matrix[i][(d + 1) : :]
                )
                tp = intersect_matrix[i][d]
                fn = FN_area[i][d]
            TP += tp
            FN += fn
            FP += fp

        TN = max(0, total_pixels - TP - FN - FP)

        # This is not used in the original implementation
        # accuracy = nan_divide(TP, all_intersecting_area)
        recall = nan_divide(TP, GT_tot_area)
        precision = nan_divide(TP, (TP + FP))
        F_factor = nan_divide(2 * (precision * recall), (precision + recall))
        true_positive_rate = nan_divide(TP, (FN + TP))
        false_positive_rate = nan_divide(FP, (FP + TN))
        false_negative_rate = nan_divide(FN, (FN + TP))
        true_negative_rate = nan_divide(TN, (FP + TN))
        shape = numpy.maximum(
            numpy.maximum(numpy.array(objects_GT.shape), numpy.array(objects_ID.shape)),
            numpy.ones(2, int),
        )
        rand_index, adjusted_rand_index = compute_rand_index_ijv(
            masks_to_ijv(objects_GT), masks_to_ijv(objects_ID), shape
        )
        results = {}

        measurement_value = {
            FTR_F_FACTOR: F_factor,
            FTR_PRECISION: precision,
            FTR_RECALL: recall,
            FTR_RAND_INDEX: rand_index,
            FTR_TRUE_POS_RATE: true_positive_rate,
            FTR_FALSE_POS_RATE: false_positive_rate,
            FTR_TRUE_NEG_RATE: true_negative_rate,
            FTR_FALSE_NEG_RATE: false_negative_rate,
            FTR_ADJUSTED_RAND_INDEX: adjusted_rand_index,
        }

        for k, v in measurement_value.items():
            results[f"{C_IMAGE_OVERLAP}_{k}"] = v

        TP_mask = subscripts(1, 1, GT_pixels, ID_pixels)
        FN_mask = subscripts(1, 0, GT_pixels, ID_pixels)
        FP_mask = subscripts(0, 1, GT_pixels, ID_pixels)
        TN_mask = subscripts(0, 0, GT_pixels, ID_pixels)

        TP_pixels = numpy.zeros((xGT, yGT))
        FN_pixels = numpy.zeros((xGT, yGT))
        FP_pixels = numpy.zeros((xGT, yGT))
        TN_pixels = numpy.zeros((xGT, yGT))

        TP_pixels = maskimg(TP_mask, TP_pixels)
        FN_pixels = maskimg(FN_mask, FN_pixels)
        FP_pixels = maskimg(FP_mask, FP_pixels)
        TN_pixels = maskimg(TN_mask, TN_pixels)
        if wants_emd:
            emd = compute_emd(
                objects_ID,
                objects_GT,
                max_points,
                decimation_method,
                max_distance,
                penalize_missing,
            )

            results[measurement_name(FTR_EARTH_MOVERS_DISTANCE)] = emd

        return results


def compute_rand_index_ijv(gt_ijv, test_ijv, shape):
    """Compute the Rand Index for an IJV matrix

    This is in part based on the Omega Index:
    Collins, "Omega: A General Formulation of the Rand Index of Cluster
    Recovery Suitable for Non-disjoint Solutions", Multivariate Behavioral
    Research, 1988, 23, 231-242

    The basic idea of the paper is that a pair should be judged to
    agree only if the number of clusters in which they appear together
    is the same.
    """
    #
    # The idea here is to assign a label to every pixel position based
    # on the set of labels given to that position by both the ground
    # truth and the test set. We then assess each pair of labels
    # as agreeing or disagreeing as to the number of matches.
    #
    # First, add the backgrounds to the IJV with a label of zero
    #
    gt_bkgd = numpy.ones(shape, bool)
    gt_bkgd[gt_ijv[:, 0], gt_ijv[:, 1]] = False
    test_bkgd = numpy.ones(shape, bool)
    test_bkgd[test_ijv[:, 0], test_ijv[:, 1]] = False
    gt_ijv = numpy.vstack(
        [
            gt_ijv,
            numpy.column_stack(
                [
                    numpy.argwhere(gt_bkgd),
                    numpy.zeros(numpy.sum(gt_bkgd), gt_bkgd.dtype),
                ]
            ),
        ]
    )
    test_ijv = numpy.vstack(
        [
            test_ijv,
            numpy.column_stack(
                [
                    numpy.argwhere(test_bkgd),
                    numpy.zeros(numpy.sum(test_bkgd), test_bkgd.dtype),
                ]
            ),
        ]
    )
    #
    # Create a unified structure for the pixels where a fourth column
    # tells you whether the pixels came from the ground-truth or test
    #
    u = numpy.vstack(
        [
            numpy.column_stack([gt_ijv, numpy.zeros(gt_ijv.shape[0], gt_ijv.dtype)]),
            numpy.column_stack(
                [test_ijv, numpy.ones(test_ijv.shape[0], test_ijv.dtype)]
            ),
        ]
    )
    #
    # Sort by coordinates, then by identity
    #
    order = numpy.lexsort([u[:, 2], u[:, 3], u[:, 0], u[:, 1]])
    u = u[order, :]
    # Get rid of any duplicate labellings (same point labeled twice with
    # same label.
    #
    first = numpy.hstack([[True], numpy.any(u[:-1, :] != u[1:, :], 1)])
    u = u[first, :]
    #
    # Create a 1-d indexer to point at each unique coordinate.
    #
    first_coord_idxs = numpy.hstack(
        [
            [0],
            numpy.argwhere((u[:-1, 0] != u[1:, 0]) | (u[:-1, 1] != u[1:, 1])).flatten()
            + 1,
            [u.shape[0]],
        ]
    )
    first_coord_counts = first_coord_idxs[1:] - first_coord_idxs[:-1]
    indexes = centrosome.index.Indexes([first_coord_counts])
    #
    # Count the number of labels at each point for both gt and test
    #
    count_test = numpy.bincount(indexes.rev_idx, u[:, 3]).astype(numpy.int64)
    count_gt = first_coord_counts - count_test
    #
    # For each # of labels, pull out the coordinates that have
    # that many labels. Count the number of similarly labeled coordinates
    # and record the count and labels for that group.
    #
    labels = []
    for i in range(1, numpy.max(count_test) + 1):
        for j in range(1, numpy.max(count_gt) + 1):
            match = (count_test[indexes.rev_idx] == i) & (
                count_gt[indexes.rev_idx] == j
            )
            if not numpy.any(match):
                continue
            #
            # Arrange into an array where the rows are coordinates
            # and the columns are the labels for that coordinate
            #
            lm = u[match, 2].reshape(numpy.sum(match) // (i + j), i + j)
            #
            # Sort by label.
            #
            order = numpy.lexsort(lm.transpose())
            lm = lm[order, :]
            #
            # Find indices of unique and # of each
            #
            lm_first = numpy.hstack(
                [
                    [0],
                    numpy.argwhere(numpy.any(lm[:-1, :] != lm[1:, :], 1)).flatten() + 1,
                    [lm.shape[0]],
                ]
            )
            lm_count = lm_first[1:] - lm_first[:-1]
            for idx, count in zip(lm_first[:-1], lm_count):
                labels.append((count, lm[idx, :j], lm[idx, j:]))
    #
    # We now have our sets partitioned. Do each against each to get
    # the number of true positive and negative pairs.
    #
    max_t_labels = reduce(max, [len(t) for c, t, g in labels], 0)
    max_g_labels = reduce(max, [len(g) for c, t, g in labels], 0)
    #
    # tbl is the contingency table from Table 4 of the Collins paper
    # It's a table of the number of pairs which fall into M sets
    # in the ground truth case and N in the test case.
    #
    tbl = numpy.zeros(((max_t_labels + 1), (max_g_labels + 1)))
    for i, (c1, tobject_numbers1, gobject_numbers1) in enumerate(labels):
        for j, (c2, tobject_numbers2, gobject_numbers2) in enumerate(labels[i:]):
            nhits_test = numpy.sum(
                tobject_numbers1[:, numpy.newaxis] == tobject_numbers2[numpy.newaxis, :]
            )
            nhits_gt = numpy.sum(
                gobject_numbers1[:, numpy.newaxis] == gobject_numbers2[numpy.newaxis, :]
            )
            if j == 0:
                N = c1 * (c1 - 1) / 2
            else:
                N = c1 * c2
            tbl[nhits_test, nhits_gt] += N

    N = numpy.sum(tbl)
    #
    # Equation 13 from the paper
    #
    min_JK = min(max_t_labels, max_g_labels) + 1
    rand_index = numpy.sum(tbl[:min_JK, :min_JK] * numpy.identity(min_JK)) / N
    #
    # Equation 15 from the paper, the expected index
    #
    e_omega = (
        numpy.sum(
            numpy.sum(tbl[:min_JK, :min_JK], 0) * numpy.sum(tbl[:min_JK, :min_JK], 1)
        )
        / N**2
    )
    #
    # Equation 16 is the adjusted index
    #
    adjusted_rand_index = (rand_index - e_omega) / (1 - e_omega)
    return rand_index, adjusted_rand_index


def compute_emd(
    src_objects,
    dest_objects,
    max_points: int,
    decimation_method: str,
    max_distance: int,
    penalize_missing: bool,
):
    """Compute the earthmovers distance between two sets of objects

    src_objects - move pixels from these objects

    dest_objects - move pixels to these objects

    returns the earth mover's distance
    """
    #
    # if either foreground set is empty, the emd is the penalty.
    #
    for angels, demons in (
        (src_objects, dest_objects),
        (dest_objects, src_objects),
    ):
        if angels.max() == 0:
            if penalize_missing:
                return numpy.sum(get_areas(demons)) * max_distance
            else:
                return 0
    if decimation_method == DM_KMEANS:
        isrc, jsrc = get_kmeans_points(src_objects, dest_objects, max_points)
        idest, jdest = isrc, jsrc
    else:
        isrc, jsrc = get_skeleton_points(src_objects, max_points)
        idest, jdest = get_skeleton_points(dest_objects, max_points)
        # TODO simplify
    src_union_binmask, dest_union_binmask = [
        get_labels_mask(numpy.array(labels_to_binmasks(x)))
        for x in (src_objects, dest_objects)
    ]
    src_weights, dest_weights = [
        get_weights(i, j, object)
        for i, j, object in (
            (
                isrc,
                jsrc,
                src_union_binmask,
            ),
            (
                idest,
                jdest,
                dest_union_binmask,
            ),
        )
    ]
    ioff, joff = [
        src[:, numpy.newaxis] - dest[numpy.newaxis, :]
        for src, dest in ((isrc, idest), (jsrc, jdest))
    ]
    c = numpy.sqrt(ioff * ioff + joff * joff).astype(numpy.int32)
    c[c > max_distance] = max_distance
    extra_mass_penalty = max_distance if penalize_missing else 0
    return centrosome.fastemd.emd_hat_int32(
        src_weights.astype(numpy.int32),
        dest_weights.astype(numpy.int32),
        c,
        extra_mass_penalty=extra_mass_penalty,
    )


def get_skeleton_points(labels: numpy.ndarray, max_points: int):
    """Get points by skeletonizing the objects and decimating"""
    total_skel = numpy.zeros(labels.shape, bool)
    # MODIFIED: Obviated get_labels step, which splits overlapping labels
    # as they are passed as arguments and we assume no support for overlaps
    colors = centrosome.cpmorphology.color_labels(labels)
    for color in range(1, numpy.max(colors) + 1):
        labels_mask = colors == color
        skel = centrosome.cpmorphology.skeletonize(
            labels_mask,
            ordering=scipy.ndimage.distance_transform_edt(labels_mask)
            * centrosome.filter.poisson_equation(labels_mask),
        )
        total_skel = total_skel | skel
    n_pts = numpy.sum(total_skel)
    if n_pts == 0:
        return numpy.zeros(0, numpy.int32), numpy.zeros(0, numpy.int32)
    i, j = numpy.where(total_skel)
    if n_pts > max_points:
        #
        # Decimate the skeleton by finding the branchpoints in the
        # skeleton and propagating from those.
        #
        markers = numpy.zeros(total_skel.shape, numpy.int32)
        branchpoints = centrosome.cpmorphology.branchpoints(
            total_skel
        ) | centrosome.cpmorphology.endpoints(total_skel)
        markers[branchpoints] = numpy.arange(numpy.sum(branchpoints)) + 1
        #
        # We compute the propagation distance to that point, then impose
        # a slightly arbitarary order to get an unambiguous ordering
        # which should number the pixels in a skeleton branch monotonically
        #
        ts_labels, distances = centrosome.propagate.propagate(
            numpy.zeros(markers.shape), markers, total_skel, 1
        )
        order = numpy.lexsort((j, i, distances[i, j], ts_labels[i, j]))
        #
        # Get a linear space of self.max_points elements with bounds at
        # 0 and len(order)-1 and use that to select the points.
        #
        order = order[numpy.linspace(0, len(order) - 1, max_points).astype(int)]
        return i[order], j[order]
    return i, j


def get_kmeans_points(src_obj, dest_obj, max_points: int):
    """Get representative points in the objects using K means

    src_obj - get some of the foreground points from the source objects
    dest_obj - get the rest of the foreground points from the destination
               objects

    returns a vector of i coordinates of representatives and a vector
            of j coordinates
    """
    from sklearn.cluster import KMeans

    ijv = numpy.vstack((src_obj.ijv, dest_obj.ijv))
    if len(ijv) <= max_points:
        return ijv[:, 0], ijv[:, 1]
    random_state = numpy.random.RandomState()
    random_state.seed(ijv.astype(int).flatten())
    kmeans = KMeans(n_clusters=max_points, tol=2, random_state=random_state)
    kmeans.fit(ijv[:, :2])
    return (
        kmeans.cluster_centers_[:, 0].astype(numpy.uint32),
        kmeans.cluster_centers_[:, 1].astype(numpy.uint32),
    )


def get_weights(i, j, labels_mask):
    """Return the weights to assign each i,j point

    Assign each pixel in the labels mask to the nearest i,j and return
    the number of pixels assigned to each i,j
    """
    #
    # Create a mapping of chosen points to their index in the i,j array
    #

    total_skel = numpy.zeros(labels_mask.shape, int)
    total_skel[i, j] = numpy.arange(1, len(i) + 1)
    #
    # Compute the distance from each chosen point to all others in image,
    # return the nearest point.
    #
    ii, jj = scipy.ndimage.distance_transform_edt(
        total_skel == 0, return_indices=True, return_distances=False
    )
    #
    # Filter out all unmasked points
    #
    ii, jj = [x[labels_mask] for x in (ii, jj)]
    if len(ii) == 0:
        return numpy.zeros(0, numpy.int32)
    #
    # Use total_skel to look up the indices of the chosen points and
    # bincount the indices.
    #
    result = numpy.zeros(len(i), numpy.int32)
    bc = numpy.bincount(total_skel[ii, jj])[1:]
    result[: len(bc)] = bc
    return result


# def measurement_name(feature, object_name_GT, object_name_ID):
def measurement_name(feature, *args):
    # name_components = [C_IMAGE_OVERLAP]
    return "_".join(
        (
            C_IMAGE_OVERLAP,
            feature,
            *args,
            # object_name_GT.value,
            # object_name_ID.value,
        )
    )


def get_areas(masks: numpy.ndarray):
    """
    Assumes mask is an flat array of integer masks
    Works for both 2D and 3D label masks.
    """
    unique_vals = numpy.unique(masks)
    unique_vals = sorted(unique_vals[unique_vals > 0])
    return numpy.array([(masks == i).sum() for i in unique_vals])


def get_labels_mask(label_masks):
    """
    Transform a set of masks into the union of all binary masks. Results in a binary 2D mask.
    """
    labels_mask = numpy.zeros(label_masks.shape[1:], bool)
    for labels in label_masks:
        labels_mask = labels_mask | (labels > 0)
    return labels_mask



"""
MeasureColocalization
=====================

**MeasureColocalization** measures the colocalization and correlation
between intensities in different images (e.g., different color channels)
on a pixel-by-pixel basis, within identified objects or across an entire
image.

Given two or more images, this module calculates the correlation &
colocalization (Overlap, Manders, Costes’ Automated Threshold & Rank
Weighted Colocalization) between the pixel intensities. The correlation
/ colocalization can be measured for entire images, or a correlation
measurement can be made within each individual object. Correlations /
Colocalizations will be calculated between all pairs of images that are
selected in the module, as well as between selected objects. For
example, if correlations are to be measured for a set of red, green, and
blue images containing identified nuclei, measurements will be made
between the following:

-  The blue and green, red and green, and red and blue images.
-  The nuclei in each of the above image pairs.

A good primer on colocalization theory can be found on the `SVI website`_.

You can find a helpful review on colocalization from Aaron *et al*. `here`_.

|

============ ============ ===============
Supports 2D? Supports 3D? Respects masks?
============ ============ ===============
YES          YES          YES
============ ============ ===============

Measurements made by this module
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

-  *Correlation:* The correlation between a pair of images *I* and *J*,
   calculated as Pearson’s correlation coefficient. The formula is
   covariance(\ *I* ,\ *J*)/[std(\ *I* ) × std(\ *J*)].
-  *Slope:* The slope of the least-squares regression between a pair of
   images I and J. Calculated using the model *A* × *I* + *B* = *J*, where *A* is the slope.
-  *Overlap coefficient:* The overlap coefficient is a modification of
   Pearson’s correlation where average intensity values of the pixels are
   not subtracted from the original intensity values. For a pair of
   images R and G, the overlap coefficient is measured as r = sum(Ri \*
   Gi) / sqrt (sum(Ri\*Ri)\*sum(Gi\*Gi)).
-  *Manders coefficient:* The Manders coefficient for a pair of images R
   and G is measured as M1 = sum(Ri_coloc)/sum(Ri) and M2 =
   sum(Gi_coloc)/sum(Gi), where Ri_coloc = Ri when Gi > 0, 0 otherwise
   and Gi_coloc = Gi when Ri >0, 0 otherwise.
-  *Manders coefficient (Costes Automated Threshold):* Costes’ automated
   threshold estimates maximum threshold of intensity for each image
   based on correlation. Manders coefficient is applied on thresholded
   images as Ri_coloc = Ri when Gi > Gthr and Gi_coloc = Gi when Ri >
   Rthr where Gthr and Rthr are thresholds calculated using Costes’
   automated threshold method.
-  *Rank Weighted Colocalization coefficient:* The RWC coefficient for a
   pair of images R and G is measured as RWC1 =
   sum(Ri_coloc\*Wi)/sum(Ri) and RWC2 = sum(Gi_coloc\*Wi)/sum(Gi),
   where Wi is Weight defined as Wi = (Rmax - Di)/Rmax where Rmax is the
   maximum of Ranks among R and G based on the max intensity, and Di =
   abs(Rank(Ri) - Rank(Gi)) (absolute difference in ranks between R and
   G) and Ri_coloc = Ri when Gi > 0, 0 otherwise and Gi_coloc = Gi
   when Ri >0, 0 otherwise. (Singan et al. 2011, BMC Bioinformatics
   12:407).

References
^^^^^^^^^^

-  Aaron JS, Taylor AB, Chew TL. Image co-localization - co-occurrence versus correlation.
   J Cell Sci. 2018;131(3):jcs211847. Published 2018 Feb 8. doi:10.1242/jcs.211847



.. _SVI website: http://svi.nl/ColocalizationTheory
.. _here: https://jcs.biologists.org/content/joces/131/3/jcs211847.full.pdf
"""

from functools import partial

import numpy
import scipy.ndimage
import scipy.stats
from centrosome.cpmorphology import fixup_scipy_ndimage_result as fix
from cp_measure.utils import labels_to_binmasks
from scipy.linalg import lstsq

M_IMAGES = "Across entire image"
M_OBJECTS = "Within objects"
M_IMAGES_AND_OBJECTS = "Both"

M_FAST = "Fast"
M_FASTER = "Faster"
M_ACCURATE = "Accurate"

"""Feature name format for the correlation measurement"""
# Modified Correlation-Pearson
F_CORRELATION_FORMAT = "Correlation_Pearson"

"""Feature name format for the slope measurement"""
F_SLOPE_FORMAT = "Correlation_Slope"

"""Feature name format for the overlap coefficient measurement"""
F_OVERLAP_FORMAT = "Correlation_Overlap"

"""Feature name format for the Manders Coefficient measurement"""
F_K_FORMAT = "Correlation_K"

"""Feature name format for the Manders Coefficient measurement"""
F_KS_FORMAT = "Correlation_KS"

"""Feature name format for the Manders Coefficient measurement"""
F_MANDERS_FORMAT = "Correlation_Manders"

"""Feature name format for the RWC Coefficient measurement"""
F_RWC_FORMAT = "Correlation_RWC"

"""Feature name format for the Costes Coefficient measurement"""
F_COSTES_FORMAT = "Correlation_Costes"


"""
thr : int or float, optional
    Set threshold as percentage of maximum intensity for the images (default 15).
    You may choose to measure colocalization metrics only for those pixels above 
    a certain threshold. Select the threshold as a percentage of the maximum intensity 
    of the above image [0-99].
    This value is used by the Overlap, Manders, and Rank Weighted Colocalization 
    measurements.

fast_costes : {M_FASTER, M_FAST, M_ACCURATE}, optional
    Method for Costes thresholding (default M_FASTER).
    This setting determines the method used to calculate the threshold for use within the
    Costes calculations. The *{M_FAST}* and *{M_ACCURATE}* modes will test candidate thresholds
    in descending order until the optimal threshold is reached. Selecting *{M_FAST}* will attempt 
    to skip candidates when results are far from the optimal value being sought. Selecting *{M_ACCURATE}* 
    will test every possible threshold value. When working with 16-bit images these methods can be extremely 
    time-consuming. Selecting *{M_FASTER}* will use a modified bisection algorithm to find the threshold 
    using a shrinking window of candidates. This is substantially faster but may produce slightly lower 
    thresholds in exceptional circumstances.
    In the vast majority of instances the results of all strategies should be identical. We recommend using 
    *{M_FAST}* mode when working with 8-bit images and *{M_FASTER}* mode when using 16-bit images.
    Alternatively, you may want to disable these specific measurements entirely 
    (available when "*Run All Metrics?*" is set to "*No*").
"""


def extract_pixels(
    pixels_1: numpy.ndarray, pixels_2: numpy.ndarray, mask: numpy.ndarray
):
    fi = pixels_1[mask]
    si = pixels_2[mask]
    labels = mask[mask]
    lrange = numpy.arange(max(labels), dtype=numpy.int32) + 1
    return fi, si, labels, lrange


def calculate_threshold(
    pixels_1: numpy.ndarray,
    pixels_2: numpy.ndarray,
    mask: numpy.ndarray,
    thr: int,
):
    # Threshold as percentage of maximum intensity of objects in each channel
    first_pixels, second_pixels, labels, lrange = extract_pixels(
        pixels_1, pixels_2, mask
    )
    tff = (thr / 100) * fix(scipy.ndimage.maximum(first_pixels, labels, lrange))
    tss = (thr / 100) * fix(scipy.ndimage.maximum(second_pixels, labels, lrange))

    combined_thresh = (first_pixels >= tff[labels - 1]) & (
        second_pixels >= tss[labels - 1]
    )
    fi_thresh = first_pixels[combined_thresh]
    si_thresh = second_pixels[combined_thresh]
    tot_fi_thr = scipy.ndimage.sum(
        first_pixels[first_pixels >= tff[labels - 1]],
        labels[first_pixels >= tff[labels - 1]],
        lrange,
    )
    tot_si_thr = scipy.ndimage.sum(
        second_pixels[second_pixels >= tss[labels - 1]],
        labels[second_pixels >= tss[labels - 1]],
        lrange,
    )
    return fi_thresh, si_thresh, tot_fi_thr, tot_fi_thr, combined_thresh


def get_correlation_pearson_ind(
    pixels_1: numpy.ndarray, pixels_2: numpy.ndarray, mask: numpy.ndarray
) -> dict[str, float]:
    fi, si, _, _ = extract_pixels(pixels_1, pixels_2, mask)
    #
    # Perform the correlation, which returns:
    # [ [ii, ij],
    #   [ji, jj] ]
    #
    corr = numpy.corrcoef((fi, si))[1, 0]
    #
    # Find the slope as a linear regression to
    # A * i1 + B = i2
    #
    coeffs = lstsq(numpy.array((fi, numpy.ones_like(fi))).transpose(), si)[0]
    slope = coeffs[0]
    return {F_CORRELATION_FORMAT: corr, F_SLOPE_FORMAT: slope}


def get_correlation_manders_fold_ind(
    pixels_1: numpy.ndarray, pixels_2: numpy.ndarray, mask: numpy.ndarray, thr: int = 15
) -> dict[str, float]:
    first_pixels, second_pixels, labels, lrange = extract_pixels(
        pixels_1, pixels_2, mask
    )
    fi_thresh, si_thresh, tot_fi_thr, tot_si_thr, combined_thresh = calculate_threshold(
        pixels_1, pixels_2, mask, thr
    )
    # Manders Coefficient
    M1 = numpy.array(
        scipy.ndimage.sum(fi_thresh, labels[combined_thresh], lrange)
    ) / numpy.array(tot_fi_thr)
    M2 = numpy.array(
        scipy.ndimage.sum(si_thresh, labels[combined_thresh], lrange)
    ) / numpy.array(tot_si_thr)

    return {
        f"{F_MANDERS_FORMAT}_1": M1[0],
        f"{F_MANDERS_FORMAT}_2": M2[0],
    }


def get_correlation_rwc_ind(
    pixels_1: numpy.ndarray, pixels_2: numpy.ndarray, mask: numpy.ndarray, thr: int = 15
) -> dict[str, float]:
    first_pixels, second_pixels, labels, lrange = extract_pixels(
        pixels_1, pixels_2, mask
    )
    fi_thresh, si_thresh, tot_fi_thr, tot_si_thr, combined_thresh = calculate_threshold(
        pixels_1, pixels_2, mask, thr
    )

    # RWC Coefficient
    RWC1 = numpy.zeros(len(lrange))
    RWC2 = numpy.zeros(len(lrange))
    [Rank1] = numpy.lexsort(([labels], [first_pixels]))
    [Rank2] = numpy.lexsort(([labels], [second_pixels]))
    Rank1_U = numpy.hstack(
        [[False], first_pixels[Rank1[:-1]] != first_pixels[Rank1[1:]]]
    )
    Rank2_U = numpy.hstack(
        [[False], second_pixels[Rank2[:-1]] != second_pixels[Rank2[1:]]]
    )
    Rank1_S = numpy.cumsum(Rank1_U)
    Rank2_S = numpy.cumsum(Rank2_U)
    Rank_im1 = numpy.zeros(first_pixels.shape, dtype=int)
    Rank_im2 = numpy.zeros(second_pixels.shape, dtype=int)
    Rank_im1[Rank1] = Rank1_S
    Rank_im2[Rank2] = Rank2_S

    R = max(Rank_im1.max(), Rank_im2.max()) + 1
    Di = abs(Rank_im1 - Rank_im2)
    weight = (R - Di) * 1.0 / R
    weight_thresh = weight[combined_thresh]

    RWC1 = numpy.array(
        scipy.ndimage.sum(fi_thresh * weight_thresh, labels[combined_thresh], lrange)
    ) / numpy.array(tot_fi_thr)
    RWC2 = numpy.array(
        scipy.ndimage.sum(si_thresh * weight_thresh, labels[combined_thresh], lrange)
    ) / numpy.array(tot_si_thr)

    return {
        f"{F_RWC_FORMAT}_1": RWC1[0],
        f"{F_RWC_FORMAT}_2": RWC2[0],
    }


def get_correlation_costes_ind(
    pixels_1: numpy.ndarray,
    pixels_2: numpy.ndarray,
    mask: numpy.ndarray,
    fast_costes: str = M_FASTER,
    thr: int = 15,
):
    # Orthogonal Regression for Costes' automated threshold
    first_pixels, second_pixels, labels, lrange = extract_pixels(
        pixels_1, pixels_2, mask
    )
    fi_thresh, si_thresh, tot_fi_thr, tot_si_thr, combined_thresh = calculate_threshold(
        pixels_1, pixels_2, mask, thr
    )
    # MODIFIED: Removed the section that combines both images
    # MODIFIED: Added dimension because algorithms expect fi,si to be 3D
    scale = infer_scale(pixels_1)
    if fast_costes == M_FASTER:
        thr_fi_c, thr_si_c = bisection_costes(
            pixels_1, pixels_2, first_pixels, second_pixels, scale, fast_costes
        )
    else:
        thr_fi_c, thr_si_c = linear_costes(
            pixels_1, pixels_2, first_pixels, second_pixels, scale
        )

    # Costes' thershold for entire image is applied to each object
    fi_above_thr = first_pixels > thr_fi_c
    si_above_thr = second_pixels > thr_si_c
    combined_thresh_c = fi_above_thr & si_above_thr
    fi_thresh_c = first_pixels[combined_thresh_c]
    si_thresh_c = second_pixels[combined_thresh_c]

    tot_fi_thr_c = numpy.zeros(len(lrange))
    tot_si_thr_c = numpy.zeros(len(lrange))

    if numpy.any(fi_above_thr):
        tot_fi_thr_c = scipy.ndimage.sum(
            first_pixels[first_pixels >= thr_fi_c],
            labels[first_pixels >= thr_fi_c],
            lrange,
        )

    if numpy.any(si_above_thr):
        tot_si_thr_c = scipy.ndimage.sum(
            second_pixels[second_pixels >= thr_si_c],
            labels[second_pixels >= thr_si_c],
            lrange,
        )

    # Costes Automated Threshold
    C1, C2 = ([0.0], [0.0])  # Cover fringe case of no pixels above threshold
    if len(fi_thresh_c) and len(si_thresh_c):
        C1 = numpy.array(
            scipy.ndimage.sum(fi_thresh_c, labels[combined_thresh_c], lrange)
        ) / numpy.array(tot_fi_thr_c)
        C2 = numpy.array(
            scipy.ndimage.sum(si_thresh_c, labels[combined_thresh_c], lrange)
        ) / numpy.array(tot_si_thr_c)

    return {
        f"{F_COSTES_FORMAT}_1": C1[0],
        f"{F_COSTES_FORMAT}_2": C2[0],
    }


def linear_costes(
    first_pixels: numpy.ndarray,
    second_pixels: numpy.ndarray,
    fi: numpy.ndarray,
    si: numpy.ndarray,
    scale_max: int = 255,
    fast_costes: str = M_FASTER,
):
    """
    Finds the Costes Automatic Threshold for colocalization using a linear algorithm.
    Candiate thresholds are gradually decreased until Pearson R falls below 0.
    If "Fast" mode is enabled the "steps" between tested thresholds will be increased
    when Pearson R is much greater than 0.
    """
    i_step = 1 / scale_max
    non_zero = (fi > 0) | (si > 0)
    xvar = numpy.var(fi[non_zero], axis=0, ddof=1)
    yvar = numpy.var(si[non_zero], axis=0, ddof=1)

    xmean = numpy.mean(fi[non_zero], axis=0)
    ymean = numpy.mean(si[non_zero], axis=0)

    z = fi[non_zero] + si[non_zero]
    zvar = numpy.var(z, axis=0, ddof=1)

    covar = 0.5 * (zvar - (xvar + yvar))

    denom = 2 * covar
    num = (yvar - xvar) + numpy.sqrt(
        (yvar - xvar) * (yvar - xvar) + 4 * (covar * covar)
    )
    a = num / denom
    b = ymean - a * xmean

    # Start at 1 step above the maximum value
    img_max = max(fi.max(), si.max())
    i = i_step * ((img_max // i_step) + 1)

    num_true = None
    fi_max = fi.max()
    si_max = si.max()

    # Initialise without a threshold
    costReg, _ = scipy.stats.pearsonr(fi, si)
    thr_fi_c = i
    thr_si_c = (a * i) + b
    while i > fi_max and (a * i) + b > si_max:
        i -= i_step
    while i > i_step:
        thr_fi_c = i
        thr_si_c = (a * i) + b
        combt = (fi < thr_fi_c) | (si < thr_si_c)
        try:
            # Only run pearson if the input has changed.
            if (positives := numpy.count_nonzero(combt)) != num_true:
                costReg, _ = scipy.stats.pearsonr(fi[combt], si[combt])
                num_true = positives

            if costReg <= 0:
                break
            elif fast_costes == M_ACCURATE or i < i_step * 10:
                i -= i_step
            elif costReg > 0.45:
                # We're way off, step down 10x
                i -= i_step * 10
            elif costReg > 0.35:
                # Still far from 0, step 5x
                i -= i_step * 5
            elif costReg > 0.25:
                # Step 2x
                i -= i_step * 2
            else:
                i -= i_step
        except ValueError:
            break
    return thr_fi_c, thr_si_c


def bisection_costes(
    first_pixels: numpy.ndarray,
    second_pixels: numpy.ndarray,
    fi: numpy.ndarray,
    si: numpy.ndarray,
    scale_max: int = 255,
    fast_costes: str = M_FASTER,
):
    """
    Finds the Costes Automatic Threshold for colocalization using a bisection algorithm.
    Candidate thresholds are selected from within a window of possible intensities,
    this window is narrowed based on the R value of each tested candidate.
    We're looking for the first point below 0, and R value can become highly variable
    at lower thresholds in some samples. Therefore the candidate tested in each
    loop is 1/6th of the window size below the maximum value (as opposed to the midpoint).
    """

    non_zero = (fi > 0) | (si > 0)
    xvar = numpy.var(fi[non_zero], axis=0, ddof=1)
    yvar = numpy.var(si[non_zero], axis=0, ddof=1)

    xmean = numpy.mean(fi[non_zero], axis=0)
    ymean = numpy.mean(si[non_zero], axis=0)

    z = fi[non_zero] + si[non_zero]
    zvar = numpy.var(z, axis=0, ddof=1)

    covar = 0.5 * (zvar - (xvar + yvar))

    denom = 2 * covar
    num = (yvar - xvar) + numpy.sqrt(
        (yvar - xvar) * (yvar - xvar) + 4 * (covar * covar)
    )
    a = num / denom
    b = ymean - a * xmean

    # Initialise variables
    left = 1
    right = scale_max
    mid = ((right - left) // (6 / 5)) + left
    lastmid = 0
    # Marks the value with the last positive R value.
    valid = 1

    while lastmid != mid:
        thr_fi_c = mid / scale_max
        thr_si_c = (a * thr_fi_c) + b
        combt = (fi < thr_fi_c) | (si < thr_si_c)
        if numpy.count_nonzero(combt) <= 2:
            # Can't run pearson with only 2 values.
            left = mid - 1
        else:
            try:
                costReg, _ = scipy.stats.pearsonr(fi[combt], si[combt])
                if costReg < 0:
                    left = mid - 1
                elif costReg >= 0:
                    right = mid + 1
                    valid = mid
            except ValueError:
                # Catch misc Pearson errors with low sample numbers
                left = mid - 1
        lastmid = mid
        if right - left > 6:
            mid = ((right - left) // (6 / 5)) + left
        else:
            mid = ((right - left) // 2) + left

    thr_fi_c = (valid - 1) / scale_max
    thr_si_c = (a * thr_fi_c) + b

    return thr_fi_c, thr_si_c


def get_correlation_overlap_ind(
    pixels_1: numpy.ndarray,
    pixels_2: numpy.ndarray,
    mask: numpy.ndarray,
    thr: int = 15,
):
    first_pixels, second_pixels, labels, lrange = extract_pixels(
        pixels_1, pixels_2, mask
    )
    _, _, _, _, combined_thresh = calculate_threshold(pixels_1, pixels_2, mask, thr)
    # Overlap Coefficient
    fpsq = scipy.ndimage.sum(
        first_pixels[combined_thresh] ** 2,
        labels[combined_thresh],
        lrange,
    )
    spsq = scipy.ndimage.sum(
        second_pixels[combined_thresh] ** 2,
        labels[combined_thresh],
        lrange,
    )
    pdt = numpy.sqrt(numpy.array(fpsq) * numpy.array(spsq))

    overlap = fix(
        scipy.ndimage.sum(
            first_pixels[combined_thresh] * second_pixels[combined_thresh],
            labels[combined_thresh],
            lrange,
        )
        / pdt
    )
    K1 = fix(
        (
            scipy.ndimage.sum(
                first_pixels[combined_thresh] * second_pixels[combined_thresh],
                labels[combined_thresh],
                lrange,
            )
        )
        / (numpy.array(fpsq))
    )
    K2 = fix(
        scipy.ndimage.sum(
            first_pixels[combined_thresh] * second_pixels[combined_thresh],
            labels[combined_thresh],
            lrange,
        )
        / numpy.array(spsq)
    )

    return {
        F_OVERLAP_FORMAT: overlap[0],
        f"{F_K_FORMAT}_1": K1[0],
        f"{F_K_FORMAT}_2": K2[0],
    }


# MODIFIED: This reproduces the behaviour of the block at
#  https://github.com/cellprofiler/CellProfiler/blob/450abdc2eaa0332cb6d1d4aaed4bf0a4b843368d/src/subpackages/core/cellprofiler_core/image/abstract_image/file/_file_image.py#L396-L405
def infer_scale(data: numpy.ndarray) -> int:
    if data.dtype in [numpy.int8, numpy.uint8]:
        scale = 255
    elif data.dtype in [numpy.int16, numpy.uint16]:
        scale = 65535
    elif data.dtype == numpy.int32:
        scale = 2**32 - 1
    elif data.dtype == numpy.uint32:
        scale = 2**32
    else:
        scale = 1

    return scale


## Define functions using skimage style
# The implementation of these correlation functions makes it irrelevant to vectorize the input.
def get_correlation_pearson(
    pixels_1: numpy.ndarray, pixels_2: numpy.ndarray, masks: numpy.ndarray
):
    return apply_correlation_fun(get_correlation_pearson_ind, pixels_1, pixels_2, masks)


def get_correlation_manders_fold(
    pixels_1: numpy.ndarray,
    pixels_2: numpy.ndarray,
    masks: numpy.ndarray,
    thr: int = 15,
):
    return apply_correlation_fun(
        get_correlation_manders_fold_ind, pixels_1, pixels_2, masks, thr=thr
    )


def get_correlation_rwc(
    pixels_1: numpy.ndarray,
    pixels_2: numpy.ndarray,
    masks: numpy.ndarray,
    thr: int = 15,
):
    return apply_correlation_fun(
        get_correlation_rwc_ind, pixels_1, pixels_2, masks, thr=thr
    )


def get_correlation_costes(
    pixels_1: numpy.ndarray,
    pixels_2: numpy.ndarray,
    masks: numpy.ndarray,
    fast_costes: str = M_FASTER,
    thr: int = 15,
):
    return apply_correlation_fun(
        get_correlation_costes_ind,
        pixels_1,
        pixels_2,
        masks,
        fast_costes=fast_costes,
        thr=thr,
    )


def get_correlation_overlap(
    pixels_1: numpy.ndarray,
    pixels_2: numpy.ndarray,
    masks: numpy.ndarray,
    thr: int = 15,
):
    return apply_correlation_fun(get_correlation_overlap_ind, pixels_1, pixels_2, masks)


# Helper functions


def apply_correlation_fun(
    corr_function,
    pixels_1: numpy.ndarray,
    pixels_2: numpy.ndarray,
    masks: numpy.ndarray,
    **kwargs,
):
    """
    Apply `corr_function` to the subsequent args and kwargs. It assumes that pixels (arrays containing images) are passed, as well as
    masks in a 2-d labels format. Any kwargs are passed to corr_functions.
    """
    results = []
    partial_corr_fun = partial(corr_function, pixels_1, pixels_2)
    for mask in labels_to_binmasks(masks):
        results.append(partial_corr_fun(mask, **kwargs))

    return {k: [item[k] for item in results] for k in results[0]}
